
import datetime
import itertools
import collections
import logging

import numpy

from .. import dataset
from .. import constants


class ACEFTS(dataset.SingleMeasurementPerFileDataset,
             dataset.ProfileDataset):
    """SCISAT Atmospheric Chemistry Experiment FTS
    """
#    basedir = "/home/gerrit/sshfs/glacier/data/1/gholl/data/ACE"
    subdir = "{year:04d}-{month:02d}"
    re = r"(?P<type>s[sr])(?P<orbit>\d{5})v(?P<version>\d\.\d)\.asc"
    _time_format = "%Y-%m-%d %H:%M:%S"
    aliases = {"CH4_profile": "CH4",
        "delta_CH4_profile": "CH4_err",
        "p": "P_pa",
        "S_CH4_profile": "CH4_SA_fake"}
    filename_fields = {"orbit": "u4", "version": "U3", "type": "U2"}
    unique_fields = {"orbit", "type", "time"}
    n_prof = "z"
    range = (5e3, 150e3)

    @staticmethod
    def read_header(fp):
        """Read header from open file

        Should be opened at the beginning.  Will advance location from
        start of header to end of header.

        :param fp: File open at beginning of header
        :returns: Dictionary with header information
        """
        head = collections.OrderedDict()
        isempty = lambda line: not line.isspace()
        for line in itertools.takewhile(isempty, fp):
            k, v = line.split("|")
            head[k.strip()] = v.strip()
        if head == {}:
            raise dataset.InvalidFileError(
                "Unable to extract header from {0.name}.  Empty?".format(fp))
        return head

    def read_single(self, f, fields="all"):
        with open(f) as fp:
            head = self.read_header(fp)

            line = fp.readline()
            while line.isspace():
                line = fp.readline()

            names = line.replace("P (atm)", "P_atm").split()
            # numpy.ma.empty fails with datetime dtype
            # https://github.com/numpy/numpy/issues/4583
            #D = numpy.ma.empty((150,),
            D = numpy.empty((150,),
                list(zip(names, ["f4"]*len(names)))
                    + [("P_pa", "f4"), ("CH4_SA_fake", "f4", (150,))])

            for (n, line) in enumerate(fp):
                # why does this not work?
                # http://stackoverflow.com/q/22865877/974555
                #D[names][n] = tuple(float(f) for f in line.split())
                try:
                    vals = tuple(float(f) for f in line.split())
                except ValueError:
                    # raise InvalidFileError instead so I can catch more
                    # narrowly higher up in the stack
                    raise dataset.InvalidFileError("Unable to read content")
                for (i, name) in enumerate(names):
                    D[name][n] = vals[i]

        # km -> m
        D["z"] *= 1e3
        D["P_pa"] = D["P_atm"] * constants.atm

        # assume error covariance matrix to be diagonal
        # and convert std. error to variance.  Errors on flagged values
        # are 0.

        val = D["CH4_err"]>0
        D["CH4_SA_fake"] = numpy.diag(D["CH4_err"]**2)
        D["CH4_SA_fake"][:, ~val] = 0
        D["CH4_SA_fake"][~val, :] = 0

        head["lat"] = float(head["latitude"])
        head["lon"] = float(head["longitude"])
        # make sure lons are in (-180, 180)
        if head["lon"] < -180:
            head["lon"] += 360
        if head["lon"] > 180:
            head["lon"] -= 360

        # for time, strip off both incomplete timezone designation and
        # decimal part (truncating it to the nearest second)
        head["time"] = datetime.datetime.strptime(
            head["date"].split(".")[0].split("+")[0], self._time_format)

        return (head, D if fields=="all" else D[fields])
        
    def get_time_from_granule_contents(self, p):
        """Get time from granule contents.

        Takes str with path, returns two datetimes
        """
        with open(p) as f:
            head = self.read_header(f)
            # cut of "+00" part, datetime defaults to UTC and having only
            # hours is contrary to any standard, so strptime cannot handle
            # it
        return tuple(datetime.datetime.strptime(
            head[m + "_time"][:-3], self._time_format)
            for m in ("start", "end"))

    def get_z(self, meas):
        try:
            return super().get_z(meas)
        except IndexError:
            pass # parent failed, continue here
        m = meas["z"]
        if m[-1] < 150: # oops, still in km
            return m * 1e3
        else:
            return m

    def flag(self, arr):
        flagged = self.combine(arr, self.related["flags"])
        flnm = self.related["flags"].aliases["flag"]
        # See e-mail Patrick 2014-06-04
        with numpy.errstate(invalid="ignore"):
            badlev = flagged[flnm]>2
            badprof = ((flagged[flnm]>=4) & (flagged[flnm]<=6)).any(1)
        logging.info("Flagging {:d}/{:d} profiles and {:d}/{:d} levels".format(
            badprof.sum(), badprof.size, badlev.sum(), badlev.size))
        arr["CH4"][badlev] = numpy.nan
        arr["CH4"][badprof, :] = numpy.nan
        arr["CH4_err"][badlev] = numpy.nan
        arr["CH4_err"][badprof, :] = numpy.nan
        arr["CH4_SA_fake"][numpy.tile(badlev[:, :, numpy.newaxis], (1,1,150))] = numpy.nan
        arr["CH4_SA_fake"][badprof, :, :] = numpy.nan
        return arr




import collections

import numpy
import scipy.interpolate

import h5py

from .. import dataset
from .. import physics
from .. import math as pamath
from ..constants import ppm as PPM, hecto as HECTO

class TansoFTSBase(dataset.ProfileDataset):
    """Applicable to both Tanso FTS versions
    """
    p_for_T_profile = numpy.array([1000, 975, 950, 925, 900, 850, 800,
        700, 600, 500, 400, 300, 250, 200, 150, 100, 70, 50, 30, 20,
        10])*HECTO # hPa -> Pa

    A_needs_converting = False
    A_needs_swapping = False

    def _read_common(self, h5f):
        """From open h5 file, read some fields common to both tanso versions
        """

        D = collections.OrderedDict()
        time_raw = h5f["scanAttribute"]["time"]
        # force UTC time
        D["time"] = numpy.array([numpy.datetime64(time_raw[i].decode('ascii')+'Z')
            for i in range(time_raw.size)], dtype="datetime64[us]")
        D["lat"] = h5f["Data"]["geolocation"]["latitude"]
        D["lon"] = h5f["Data"]["geolocation"]["longitude"]
        D["z0"] = h5f["Data"]["geolocation"]["height"]
        D["T"] = h5f["scanAttribute"]["referenceData"]["temperatureProfile"]
        D["h2o"] = h5f["scanAttribute"]["referenceData"]["waterVaporProfile"][...]*PPM
        D["p0"] = h5f["scanAttribute"]["referenceData"]["surfacePressure"][...] * HECTO
        D["id"] = h5f["scanAttribute"]["scanID"]

        return D

class TansoFTSv10x(dataset.MultiFileDataset, TansoFTSBase):
    """Tanso FTS v1.00, 1.01
    """

    re = r"GOSATTFTS(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})_02P02TV010[01]R\d{6}[0-9A-F]{5}\.h5"
    aliases = {"CH4_profile": "ch4_profile",
               "ak": "ch4_ak",
               "delta_CH4_profile": "ch4_profile_e"}
    n_prof = "p"
    # For TANSO collocating with PEARL, 50% of profiles have > 40%
    # sensitivity between 5.3 km and 9.7 km.
    #
    # For TANSO collocating with ACE, 20% of profiles have > 30%
    # sensitivity between 5.2 km and 9.5 km.
    range = (5e3, 10e3)

    # NOTE: For Tanso FTS v1.0x, there are THREE pressure profiles:
    #
    # - CH4 has 23 levels: 1165.9, 857.7, 735.6, 631.0, 541.2, 464.2,
    #   398.1, 341.4, 287.3, 237.1, 195.7, 161.6, 133.4, 100.0, 75.0, 51.1,
    #   34.8, 23.7, 16.2, 10.0, 5.6, 1.0, 0.1 hPa (GOSAT Product Description)
    #
    # - For each retrieval, the average pressure in each of 22 layers is
    #   stored (this one varies)
    #
    # - For temperature and water vapour, there are 21 vertical levels:
    #   1000, 975, 950, 925, 900, 850, 800, 700, 600, 500, 400, 300, 250,
    #   200, 150, 100, 70, 50, 30, 20, 10 hPa.
    #
    # To get CH4 on z-grid rather than p-grid, need to convert p-grid to
    # z-grid.  This needs temperature and water vapour profiles.  Either
    # interpolate *and* extrapolate z-grid onto retrieval grid, or
    # interpolate retrieval *and* averaging kernel onto T/Q-grid.  Prefer
    # the former because ak-interpolation I found is linear which is OK
    # for z but not for p?
    #

    # implementation of abstract methods
    def _read(self, path, fields="all"):
        D = collections.OrderedDict()
        with h5py.File(path, 'r') as h5f:
            D.update(self._read_common(h5f))
            D["ch4_profile"] = (h5f["Data"]["originalProfile"]
                ["CH4Profile"][:]*PPM)
            D["ch4_profile_e"] = (h5f["Data"]["originalProfile"]
                ["CH4ProfileError"][:]*PPM)
            D["ch4_ak"] = h5f["Data"]["originalProfile"]["CH4AveragingKernelMatrix"]
            D["p"] = h5f["Data"]["originalProfile"]["pressure"][...]*HECTO
            D["z0"] = h5f["Data"]["geolocation"]["height"]
            A = numpy.empty(shape=D["time"].size,
                dtype=[(k, D[k].dtype, D[k].shape[1:]) for k in D.keys()])
            for k in D.keys():
                A[k] = D[k][:]
        return A if fields=="all" else A[fields]

    def get_z(self, obj, field=None):
        if field is None:
            try:
                return super().get_z(obj)
            except IndexError:
                pass # parent failed, continue here
        elif field != "T":
            raise ValueError("No special case for {}".format(field))
        # See comment near top of this class; different p-grids, different
        # z-grids.  Want z-grid corresponding to CH4-profile.
        p_for_T = self.p_for_T_profile # down up to 1 kPa
        p_for_CH4 = obj["p"] # down to ~50 Pa
        p_CH4_valid = numpy.isfinite(p_for_CH4)
        p_CH4_valid[p_CH4_valid] = p_for_CH4[p_CH4_valid] > 0
        p_for_CH4 = p_for_CH4[p_CH4_valid]
        p_T_valid = numpy.isfinite(p_for_T)
        p_T_valid[p_T_valid] = p_for_T[p_T_valid] > 0
        p_for_T = p_for_T[p_T_valid]
        p_valid = p_T_valid if field == "T" else p_CH4_valid
        # inter- and extra-polate T and h2o onto the p_for_CH4 grid
        # This introduces an extrapolation error but between 1 kPa and 50
        # Pa I don't really care, should be insignificant.

        # highest level is lowest pressure i.e. lower bound, and vice
        # versa.  Therefore reserve p, f(p) so p is going up (increasing
        # pressure)
        xb = min(p_for_CH4.min(), p_for_T.min())
        # including obj["p0"] here guarantees reference pressure/altitude
        # always included in profile, satisfying Patricks algorithm
        # (p2z_hydrostatic)
        # The purpose of this interpolation is to find temperature and
        # water vapour mixing ratio in order to convert pressure to
        # elevation.  Even setting water vapour mixing ratio completely
        # to 0 leads to sub-metre differences in elevation calculations.
        # This means that a "1-D spline" (i.e. linear interpolation in T
        # and log(vmr_h2o), k=1) is acceptable.
        xe = max(p_for_CH4.max(), p_for_T.max(), obj["p0"])
        tck_T = scipy.interpolate.splrep(p_for_T[::-1], obj["T"][::-1],
            xb=xb,
            xe=xe,
            k=1)
        tck_logh2o = scipy.interpolate.splrep(p_for_T[::-1], numpy.log(obj["h2o"][::-1]),
            xb=xb,
            xe=xe,
            k=1)

        # extrapolate, add extra pressure points.  Make sure that
        # p0 >= p_grid[0], also include 100 kPa as this was in the
        # original grid.  For completeness, merge both pressure grids,
        # then "unpack" later.  Sort by index so I can later extract the
        # parts that were belonging to p_for_CH4.
        #
        # FIXME: Do I really need to "unpack" it again?  Or just stick
        # with the merged grid?!
        p_full = []
        p_full.append(obj["p0"])
        p_for_T_i0 = len(p_full)
        p_full.extend(p_for_T)
        p_for_T_i1 = len(p_full)
        p_for_ch4_i0 = len(p_full)
        p_full.extend(p_for_CH4)
        p_for_ch4_i1 = len(p_full)
        p_full = numpy.array(p_full, dtype="f4")
        p_inds = numpy.argsort(p_full)[::-1] # high to low pressure

        # masked arrays go wrong within the p2z algorithm...
        # rather do masking "by hand" :(
        p_full_sorted = p_full[p_inds]
        ok = (numpy.isfinite(p_full_sorted) &
              (p_full_sorted > 0))
#        p_full_sorted = numpy.ma.masked_less(p_full_sorted, 0)
#        T_retgrid = numpy.ma.masked_array(
        T_retgrid = numpy.zeros_like(p_full_sorted)
        T_retgrid.fill(numpy.nan)
        T_retgrid[ok] = scipy.interpolate.splev(
            p_full_sorted[ok], tck_T, ext=1)#,
#            p_full_sorted.mask)
#        q_retgrid = numpy.ma.masked_array(
        logq_retgrid = numpy.zeros_like(p_full_sorted)
        logq_retgrid.fill(numpy.nan)
        logq_retgrid[ok] = scipy.interpolate.splev(
            p_full_sorted[ok], tck_logh2o, ext=1)#,
        q_retgrid = numpy.zeros_like(p_full_sorted)
        q_retgrid.fill(numpy.nan)
        q_retgrid[ok] = numpy.exp(logq_retgrid[ok])
#            p_full_sorted.mask)

        z_extp = numpy.zeros(shape=p_full_sorted.shape, dtype="f4")
        z_extp.fill(numpy.nan)
        ok = (ok &
              numpy.isfinite(T_retgrid) &
              numpy.isfinite(q_retgrid))
        ok[ok] = (ok[ok] & 
              (T_retgrid[ok] > 0) &
              (q_retgrid[ok] > 0))
        z_extp[ok] = physics.p2z_hydrostatic(
            p_full_sorted[ok], T_retgrid[ok], q_retgrid[ok],
            obj["p0"], obj["z0"], obj["lat"], -1)
        
        # return only elements belonging to p_for_CH4
        # this yields the "inverse" sorting indices to get back the
        # original order, but only for those that were part of p_for_CH4
        p_i_was_CH4 = ((p_inds>=p_for_ch4_i0) & (p_inds<p_for_ch4_i1)).nonzero()[0]
        p_i_was_T   = ((p_inds>=p_for_T_i0)   & (p_inds<p_for_T_i1)).nonzero()[0]

        z_ret = numpy.zeros(dtype="f4", 
            shape=(p_T_valid if field=="T" else p_CH4_valid).shape)
        z_ret[p_valid] = z_extp[p_i_was_T if field=="T" else p_i_was_CH4]
        z_ret[~p_valid] = numpy.nan
        return z_ret
        #return z_extp[p_i_was_CH4]#.data # no masked array...
#        inv_inds[-len(p_for_CH4):]]

#        z_for_T = physics.p2z_hydrostatic(
#            self.p_for_T_profile, obj["T"], obj["h2o"], 
#            obj["p0"], obj["z0"], obj["lat"], -1)
    
    def get_z_for(self, obj, field):
        return self.get_z(obj, field)

    def get_time_from_granule_contents(self, p):
        M = self.read(p, fields={"time"})
        return (M["time"].min(), M["time"].max())
            

class TansoFTSv001(dataset.SingleFileDataset, TansoFTSBase):
    # Retained for historical reasons

#    start_date = datetime.datetime(2010, 3, 23, 2, 24, 54, 210)
#    end_date = datetime.datetime(2010, 10, 31, 20, 34, 50, 814)
#    srcfile = ("/home/gerrit/sshfs/glacier/data/1/gholl/data/1403050001/"
#               "GOSATTFTS20100316_02P02TV0001R14030500010.h5")
#    name = "GOSAT Tanso FTS"

    p_for_interp_profile = numpy.array([1000, 700, 500, 300, 100, 50,
        10])*HECTO
    aliases = {"CH4_profile": "ch4_profile_raw"}

    @classmethod
    def vmr_to_column_density(cls, data):
        """Convert VMR to column density

        Currently hardcoded to CH4 profiles and the dtype by this classes'
        read method.

        Returns number column density and its error

        Note that a small error may be introduced if profiles do not go
        down to the reference provided by (z0, p0), because the p2z
        function needs a reference inside the p-profile.  This is
        mitigated by adding a dummy level with a dummy temperature, then
        removing this after the z-profile is calculated.  This has a minor
        (<0.1%) effect on the remaining levels.
        """
        # temperature interpolated on each pressure grid
        ncd = numpy.empty(shape=data.shape, dtype="f8")
        ncd_e = numpy.empty_like(ncd)

        for n in range(data.size):
            if data["T"].mask.size == 1:
                T = data["T"][n, ::-1].data
            else:
                T = data["T"][n, ::-1].data[~data["T"][n, ::-1].mask]

            if data["h2o"].mask.size == 1:
                h2o = data["h2o"][n, ::-1].data
            else:
                h2o = data["h2o"][n, ::-1].data[~data["h2o"][n, ::-1].mask]

            T_interpolator = scipy.interpolate.interp1d(
                cls.p_for_T_profile[::-1], T)# should be increasing

            h2o_interpolator = scipy.interpolate.interp1d(
                cls.p_for_T_profile[::-1], h2o)
            
            p = data["p_raw"][n, :].data[~data["p_raw"][n, :].mask]*HECTO

            T_interp = T_interpolator(p)
            h2o_interp = h2o_interpolator(p)

            nd = physics.vmr2nd(
                data["ch4_profile_raw"][n, :].data[
                    ~data["ch4_profile_raw"][n, :].mask],#*PPM,
                T_interp, p)

            nd_e = physics.vmr2nd(
                data["ch4_profile_raw_e"][n, :].data[
                    ~data["ch4_profile_raw_e"][n, :].mask],#*PPM,
                T_interp, p)

            #z = physics.p2z_oversimplified(p)
            if data["p0"][n] > p[0]:
                # need to add dummy p[0] < p to satisfy p2z
                #
                # this is inaccurate, but has little effect on the
                # *relative* location of other levels (<1 m), so it is an
                # acceptable source of error
                dummy = True
                pp = numpy.r_[data["p0"][n], p]
                Tp = numpy.r_[287, T_interp]
                # although putting h2o-surface to 0 is an EXTREMELY
                # bad approximation, it /still/ doesn't matter for the
                # relative thicknesses higher up
                h2op = numpy.r_[h2o_interp[0], h2o_interp]
            else:
                dummy = False
                (pp, Tp, h2op) = (p, T_interp, h2o_interp)
            #z = physics.p2z_hydrostatic(p, T_interp, h2o_interp,
            z = physics.p2z_hydrostatic(pp, Tp, h2op*PPM,
                p0=data["p0"][n],
                z0=data["z0"][n],
                lat=data["lat"][n])[dummy:]

            ncd[n] = pamath.integrate_with_height(z, nd)
            ncd_e[n] = pamath.integrate_with_height(z, nd_e)

        return (ncd, ncd_e)


    # implementation of abstract methods

    def _read(self, path=None, fields="all"):
        """Read Tanso FTS granule.  Currently hardcoded for CH4 raw&interp.
        
        """
        if path is None:
            path = self.srcfile

        with h5py.File(path, 'r') as h5f:
            D = collections.OrderedDict()
            D.update(self._read_common(h5f))
            p = h5f["Data"]["interpolatedProfile"]["pressure"][:]
            p *= HECTO # Pa -> hPa
            if self.p_for_interp_profile is not None:
                if not (self.p_for_interp_profile == p).all():
                    raise ValueError("Found inconsistent pressure"
                            " profiles!")
            else:
                self.p_for_interp_profile = p
            #D["p"] = h5f["Data"]["originalProfile"]["pressure"]
            D["ch4_profile_interp"] = (h5f["Data"]["interpolatedProfile"]
                ["CH4Profile"][:]*PPM)
            D["ch4_profile_raw"] = (h5f["Data"]["originalProfile"]
                ["CH4Profile"][:]*PPM)
            D["ch4_profile_raw_e"] = (h5f["Data"]["originalProfile"]
                ["CH4ProfileError"][:]*PPM)
            D["p_raw"] = h5f["Data"]["originalProfile"]["pressure"]

            A = numpy.empty(shape=time_raw.size,
                dtype=[(k, D[k].dtype, D[k].shape[1:]) for k in D.keys()])
            for k in D.keys():
                A[k] = D[k][:]
            # Unfortunately, masked arrays are buggy
#            A = A.view(numpy.ma.MaskedArray)
#            for k in {"ch4_profile_interp", "ch4_profile_raw",
#                      "ch4_profile_raw_e", "p_raw", "T"}:
#                A.mask[k][A.data[k]<0] = True

        return A if fields=="all" else A[fields]

    def get_z(self, meas): 
        try:
            return super().get_z(meas)
        except IndexError:
            pass # parent failed, continue here
#            if data["T"].mask.size == 1:
#                T = data["T"][n, ::-1].data
#            else:
#                T = data["T"][n, ::-1].data[~data["T"][n, ::-1].mask]
#
#            if data["h2o"].mask.size == 1:
#                h2o = data["h2o"][n, ::-1].data
#            else:
#                h2o = data["h2o"][n, ::-1].data[~data["h2o"][n, ::-1].mask]


        # Can't use masked arrays due to bug:
        # https://github.com/numpy/numpy/issues/2972
        # Use workaround instead.
        if isinstance(meas, numpy.ma.core.MaskedArray):
            meas = meas.data

        valid = meas["p_raw"] > 0

        T_interpolator = scipy.interpolate.interp1d(
            self.p_for_T_profile[::-1],
            meas["T"][::-1])# should be increasing

        h2o_interpolator = scipy.interpolate.interp1d(
            self.p_for_T_profile[::-1],
            meas["h2o"][::-1])
        
#            p = data["p_raw"][n, :].data[~data["p_raw"][n, :].mask]*HECTO
        p = meas["p_raw"] * HECTO
        p = p[valid]

        T_interp = T_interpolator(p)
        h2o_interp = h2o_interpolator(p)

#            nd = physics.vmr2nd(
#                data["ch4_profile_raw"][n, :].data[
#                    ~data["ch4_profile_raw"][n, :].mask]*PPM,
#                T_interp, p)
#
#            nd_e = physics.vmr2nd(
#                data["ch4_profile_raw_e"][n, :].data[
#                    ~data["ch4_profile_raw_e"][n, :].mask]*PPM,
#                T_interp, p)

        #z = physics.p2z_oversimplified(p)
        if meas["p0"] > p[0]:
            # need to add dummy p[0] < p to satisfy p2z
            #
            # this is inaccurate, but has little effect on the
            # *relative* location of other levels (<1 m), so it is an
            # acceptable source of error
            dummy = True
            pp = numpy.hstack((meas["p0"], p))
            Tp = numpy.hstack((287, T_interp))
            # although putting h2o-surface to a low value is an EXTREMELY
            # bad approximation, it /still/ doesn't matter for the
            # relative thicknesses higher up
            h2op = numpy.hstack((h2o_interp[0], h2o_interp))
        else:
            dummy = False
            (pp, Tp, h2op) = (p, T_interp, h2o_interp)
        #z = physics.p2z_hydrostatic(p, T_interp, h2o_interp,
        z = physics.p2z_hydrostatic(pp, Tp, h2op*PPM,
            p0=meas["p0"],
            z0=meas["z0"],
            lat=meas["lat"])[dummy:]
        return z




#!/usr/bin/python
# coding: utf-8

"""Functions related to pyatmlab itself, or for internal use
"""

import sys
import pkg_resources

def get_version(path_to_changelog):
    """Obtain version number from latest ChangeLog entry
    
    :param path_to_changelog: Path to ChangeLog file
    :type path_to_changelog: String-like
    :returns: String with version number major.minor.micro
    """

    with open(path_to_changelog, 'r', encoding='utf-8') as f:
        # should be on third line
        f.readline()
        f.readline()
        line = f.readline()
        return line[line.find("pyatmlab")+9:].strip().replace('-', '.')

def expanddoc(f):
    """Function decorator applying str.format-replacement on function docstring

    :param f: function to alter docstring at
    :returns: function
    """
    f.__doc__ = f.__doc__.format(**vars(sys.modules[f.__module__]))
    return f

def get_full_version():
    """Get version number from pkg_resources
    """

    return pkg_resources.get_distribution("pyatmlab").version



#!/usr/bin/env python

"""Describe batches of data: atmospheric db, scattering db, etc.

This module contains a class with functionality to read atmospheric data
from a variety of formats and write it to a variety of formats.

Mostly obtained from PyARTS.
"""

import sys
import abc
import copy
import random
import itertools
import pickle
import pathlib
import logging
import lzma

import numpy
import datetime
now = datetime.datetime.now

import numpy.lib.recfunctions
#import numpy.ma
import scipy.io

import matplotlib.mlab # contains PCA class
#import matplotlib.pyplot
#import matplotlib.cm

import progressbar

from . import tools
from . import stats
from . import config

class AtmosphericDatabase:
    """Represents an atmospheric database

    Apart from the default constructor, those constructors may be useful:

    - :func:`AtmosphericDatabase.from_evans07`
    - :func:`AtmosphericDatabase.from_evans12`

    Attributes:

    - ``data``
    - ``instrument``
    - ``name``
    """

    def __init__(self, **kwargs):
        """Constructor for atmospheric database

        Takes only keyword arguments, where each keyword ends up as an
        instance attribute, i.e.
        AtmosphericDatabase(instrument=pyatmlab.instruments.ici).
        """

        for (k, v) in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def from_evans07(cls, dbfile, instrument=None):
        """Read from db in style for Evans study in 2006-2007

        :param dbfile: Path to database-file
        :param instrument: Optional argument, pass instrument.
        """

        with open(dbfile, 'r') as fp:
            
            ncases = int(fp.readline().strip().split()[0])
            fp.readline() # ncloudpar ntemplay nrhlay nchan nztab nmutab ntinfo nuinfo
            fp.readline() # RH layers tops / bottoms
            chan_names = fp.readline().split("=")[0].strip().split()
            fp.readline() # ztab
            fp.readline() # mutab
            fp.readline() # header line

            # pre-allocate
            n_bt = len(chan_names)
            Z = numpy.zeros(ncases, dtype=cls.get_dtype(n_bt))

            i = 0
            while i<ncases:
                line = fp.readline()
                (iwp, dme, zmed, *rest) = (float(x) for x in line.strip().split())
                fp.readline() # empty line
                bt = numpy.array([float(x) for x in fp.readline().strip().split()])
                Z["BT"][i] = bt
                Z["Dme"][i] = dme
                Z["Zmed"][i] = zmed
                Z["IWP"][i] = iwp
                i += 1
                if i%(ncases//5) == 0:
                    print(now(), "done", "{}/{}".format(i, ncases))

        return cls(data=Z, instrument=instrument, name="colorado2007")

    @classmethod
    def from_evans12(cls, dbfile, instrument=None, profile=False,
                   radar=False, maxlen=None, ext=False):
        """Read from db in style for Evans study in 2012

        :param dbfile: Path to database-file
        :param instrument: Optional argument, pass instrument.
        :param profile: Include profile information.  Defaults to False.
        :param radar: Include radar information.  Defaults to False.
        :param int maxlen: Optional maximum length of db to read.
        :param bool ext: Read extended version, with manual additions by
            Gerrit 2014-01-16 (i.e. also outputting non-retrieved
            quantities used in the RTM)
        """

        print(now(), "Reading", dbfile)
        fp = open(dbfile, 'r')

        # skip and interpret header
        fp.readline() # header line
        ncases = int(fp.readline().strip().split()[0])
        maxlen = maxlen or ncases
        fp.readline() # no. of dimensions of independent Gaussian prior space
        fp.readline() # no. of elements in observation vector
        nchans = int(fp.readline().strip().split()[0])
        chan_names = fp.readline().split()
#            print(now(), "compare:", self.instrument.channel_string(), chan_names)

        # no. values per channel
        chan_lengths = numpy.array([int(c)
                                    for c in fp.readline().strip().split()])
        fp.readline() # no. of viewing angles
        fp.readline() # cos for each viewing angle
        fp.readline() # random seed
        fp.readline() # no. retrieved quantities
        fp.readline() # no. integrated cloud params
        n_layer_rh = int(fp.readline().strip().split()[0]) # no. retrieved humidity levels
        fp.readline() # height per humidity layer
        n_layer_ice = int(fp.readline().strip().split()[0]) # no. retrieved ice cloud layers
        (ice_bottom, ice_top) = (float(c) for c in
                                 fp.readline().split('!')[0].strip().split()) # bottom and top height ice cloud region
        ice_axis = numpy.linspace(ice_bottom, ice_top, n_layer_ice)
        fp.readline() # header line
        if ext:
            n_aux = int(fp.readline().strip().split()[0]) # no. aux

        chan_is_radar = numpy.array(["radar" in x.lower()
                                     for x in chan_names])

        # construct dtype
        #
        # Will have two dtypes:
        # - one where all channels are named
        # - one with a single large dtype for all radiances

        dt = [("IWP", numpy.float32),
              ("Dme", numpy.float32),
              ("Zmed", numpy.float32)]

        dt.extend([("shape", 
                    [("plateagg", numpy.float32),
                     ("sphragg", numpy.float32),
                     ("snowagg", numpy.float32),
                     ("hail", numpy.float32)])])

        if profile:
            dt.append(("IWC", numpy.float32, n_layer_ice))
            dt.append(("Dme_prof", numpy.float32, n_layer_ice))
            dt.append(("RH", numpy.float32, n_layer_rh))

        if ext:
            n_per_aux = n_aux//4
            dt.append(("height", numpy.float32, n_per_aux))
            dt.append(("temp", numpy.float32, n_per_aux))
            dt.append(("pres", numpy.float32, n_per_aux))
            dt.append(("r", numpy.float32, n_per_aux))

        dt_alt = copy.copy(dt)

        dt.extend([(nm, numpy.float32, ln)
                   for (nm, ln) in zip(chan_names, chan_lengths)])

        dt_alt.append(("BT", numpy.float32,
                        numpy.count_nonzero(~chan_is_radar)))
        dt_alt.extend([(nm, numpy.float32, ln)
                       for (nm, ln, israd) in zip(chan_names,
                                                  chan_lengths,
                                                  chan_is_radar)
                       if israd])

#        if radar:
#            dt.extend([(nm, numpy.float32, chanlen)
#                       for (nm, chanlen, israd)
#                       in zip(chan_names, chan_lengths, chan_is_radar)
#                       if israd])

        # get index in measurement array for each measurement
        edges = numpy.r_[0, chan_lengths.cumsum()]
        locations = {chan_names[k]: slice(edges[k], edges[k+1])
                     for k in range(chan_lengths.size)}

#        passive = numpy.array([ln
#                               for ln in chan_lengths
#                               if not "radar" in nm.lower()]

        #n_bt = nchans - sum(["radar" in x.lower() for x in chan_names])

        # pre-allocate
        data = numpy.empty(maxlen, dtype=dt)
        data_alt = data.view(dtype=dt_alt)

        # use while, not for-loop, because I read several lines per
        # iteration
        i = 0
        while i<maxlen:
            line = fp.readline()
            if not line: # empty means EOF
                break
#            if i%nth != 0: # allow for reading less
#                continue
            (rvcheck, IWP, Dme, Zmed,
             plateagg, sphragg, snowagg, hail,
             meltLWP, cldLWP) = (float(x) for x in line.strip().split())
            line = fp.readline()
            RH = numpy.array([float(x) for x in line.strip().split()])
            line = fp.readline()
            IWC = numpy.array([float(x) for x in line.strip().split()])
            line = fp.readline()
            Dmeprof = numpy.array([float(x) for x in line.strip().split()])
            line = fp.readline()
            measurements = numpy.array([float(x)
                                        for x in line.strip().split()])
            if ext:
                line = fp.readline()
                aux = numpy.array([float(x) for x in
                                   line.strip().split()])
#            BT = numpy.array(measurements[~chan_is_radar])
#            radar = numpy.array(measurements[chan_is_radar])
    #        radar_integrated = numpy.array(measurements[9])
    #        radar_prof = numpy.array(measurements[10:])
            #print "measurements:", BT
            #print "IWP:", IWP
            #print "IWC:", IWC
            data["IWP"][i] = IWP
            data["Dme"][i] = Dme
            data["Zmed"][i] = Zmed
#            data["BT"][i] = BT
            data["shape"][i]["plateagg"] = plateagg
            data["shape"][i]["sphragg"] = sphragg
            data["shape"][i]["snowagg"] = snowagg
            data["shape"][i]["hail"] = hail

            for nm in chan_names:
                data[nm][i] = measurements[locations[nm]]

            if profile:
                data["IWC"][i] = IWC
                data["Dme_prof"][i] = Dmeprof
                data["RH"][i] = RH

            if ext:
                data["height"][i] = aux[0*n_per_aux:1*n_per_aux]
                data["temp"][i] = aux[1*n_per_aux:2*n_per_aux]
                data["pres"][i] = aux[2*n_per_aux:3*n_per_aux]
                data["r"][i] = aux[3*n_per_aux:4*n_per_aux]
#            if radar:
#                for radfield in chan_is_radar.nonzero()[0]:
#                    nm = chan_names[radfield]
#                    data[nm][i] = measurements
            i += 1
            if i%(maxlen//8) == 0:
                print(now(), "done", "{}/{}, {}%".format(i, maxlen,
                                                         (i/maxlen)*100))

        return cls(data=data_alt, instrument=instrument,
                   name="colorado2012", ice_axis=ice_axis)


    @tools.validator
    def __getitem__(self, key: str):
        """Get field from data.

        :param str key: Field name (from self.data["dtype"]).
        :returns: ndarray, view from self.data
        """
        return self.data[key]

    @tools.validator
    def __setitem__(self, key: str, val: numpy.ndarray):
        """Add field to data or set existing field to data.

        :param str key: Field name
        :param ndarray val: Field value.  Must match dimensions of
            self.data.  Note that this will be added with dtype([key,
            val.dtype]).
        """

        if key in self.data.dtype.names:
            self.data[key] = val
        else:
            prim = self.data
            sec = val.view(dtype=[(key, val.dtype)])
            self.data = numpy.lib.recfunctions.merge_arrays(
                (prim, sec)).view(dtype=(prim.dtype.descr + sec.dtype.descr))
            

    # calculating different statistics

#     def stats_IWP_Dme(self, func, minlen=20, *args, **kwargs):
#         """Calculate a statistic per logIWP/Dme
# 
#         :param func: Statistical function to apply for each bin
#         :param minlen: Mask statistic if number of values per bin are less
#             than this.  Defaults to 20.
# 
#         All remaining arguments are passed on to
#         :func:`tools.filter_array`.
# 
#         Returns (xbin, ybin, stats)
#         """
# 
#         xbin = numpy.linspace(-1, 4, 20)
#         ybin = numpy.linspace(0, 650, 20)
#         #pos = self.data["IWP"] > 0
#         minlen = 20
#         nchans = self.data["BT"].shape[1]
#         filler = numpy.empty(nchans)
#         filler.fill(numpy.nan)
#         filt = tools.filter_array(self.data, IWP=(1e-5, 1e10),
#                                   *args, **kwargs)
#         binned_data = tools.bin2d(self.data[filt], "IWP", xbin, "Dme", ybin,
#                                   filter1=numpy.log10)
#         binned_stats = (numpy.array([
#             [func(b["BT"]) if len(b["BT"])>minlen else filler for b in r]
#                 for r in binned_data]))
#         return (xbin, ybin, numpy.ma.masked_invalid(binned_stats))
# 
#     def stats3d(self, func, minlen=20):
#         """Calculate statistic for 3-D bins per IWP, Dme, Zmed
#         """
# 
#         bin_iwp = numpy.linspace(-1, 4, 14)
#         bin_dme = numpy.linspace(0, 650, 15)
#         bin_zmed = numpy.linspace(0, 20, 16)
# 
#         binned = tools.bin3d(self.data,
#            "IWP", bin_iwp,
#            "Dme", bin_dme,
#            "Zmed", bin_zmed,
#            filter1=numpy.log10)
# 
#         nchans = self.data["BT"].shape[1]
#         filler = numpy.empty(nchans)
#         filler.fill(numpy.nan)
# 
#         stats3d = numpy.array([[
#             [func(e["BT"]) if len(e["BT"])>minlen else filler for e in r]
#                 for r in c]
#                     for c in binned])
# 
#         return (bin_iwp, bin_dme, bin_zmed, stats3d)
#  
#     def shape_per_iwp(self):
#         """Get distribution of shapes as a function of IWP
#         """
# 
#         bins_iwp = numpy.linspace(-1, 4, 14)
#         filt = self.data["IWP"] > 0
#         binned = tools.bin(numpy.log10(self.data["IWP"][filt]),
#                            self.data[filt], bins_iwp)
#         shape_dist = numpy.empty_like(bins_iwp, dtype = self.data.dtype["shape"])
#         for shape in self.data.dtype["shape"].names:
#             shape_dist[shape] = [v["shape"][shape].mean() for v in binned]
#         return shape_dist
# 
#     # visualisations
# 
#     def plot_stat_IWP_Dme(self, statfunc, ch,
#                           minlen=20, statname=None, ax=None,
#                           *args, **kwargs):
#         """Visualise statistic IWP Dme in a hist2d.
# 
#         Requires instrument to be defined.
# 
#         Input:
#             - statistic (callable)
#             - channel
#             - minlen, statname, ax
#             - all other passed to tools.filter_array
# 
#         Optional input:
# 
#             - minlen
#             - statname
#             - axes object
# 
#         Returns:
# 
#             fig, ax, pcolormesh, colorbar
#         """
# 
#         if statname is None:
#             statname = statfunc.__name__
# 
#         if ax is None:
#            fig = matplotlib.pyplot.figure()
#            ax = fig.add_subplot(1, 1, 1,
#                xlabel="${}^{10}$log IWP [g/m^2]",
#                xscale="log",
#                ylabel="Dme [µm]",
#                title="{2} {0.name} {0.instrument.name} {1:s} GHz".format(
#                    self,
#                    self.instrument.channels[ch-1].get_chanstr(full=True),
#                    statname))
#        else:
#            fig = ax.figure
#
#        (xbin, ybin, stat) = self.stats_IWP_Dme(statfunc, minlen=minlen,
#                                *args, **kwargs)
#        cmap = matplotlib.cm.Spectral_r
#        cmap.set_bad(color="white", alpha=None)
#        #cmap.set_over(color="black")
#        #cmap.set_under(color="cyan")
#        pm = ax.pcolormesh(10**xbin, ybin, stat[..., ch-1].T, cmap=cmap)
#        #pm.set_clim([0, 15])
#        cb = fig.colorbar(pm)
#        #cb.set_label("interquantile range [K]")
#        
#        return (fig, ax, pm, cb)
#
#    def plot_lines_stat3d(self, statfunc):
#        (bins_iwp, bins_dme, bins_zmed, stat) = self.stats3d(statfunc)
#        isfin = numpy.isfinite(stat[..., 0])
#        fin_in_iwp = isfin.sum(2).sum(1)
#        fin_in_dme = isfin.sum(2).sum(0)
#        fin_in_zmed = isfin.sum(1).sum(0)
#        for (iwp_i, iwp) in enumerate(bins_iwp):
#            if fin_in_iwp[iwp_i] > 10:
#                # one zmed-line per dme for this iwp
#                plot(bins_dme, stat[iwp_i, ..., ch])
#                # one dme-line per zmed for this iwp
#                plot(bins_zmed, stat[iwp_i, ..., ch].T)
##            for (dme_i, dme) in bins_dme:
##                for (zmed_i, zmed) in bins_zmed:
##                    pass
#           

    # writing out the results in different ways

    def write_evans_obs(self, fn, sel=None):
        """Write as evans-obs file

        :param fn: File to write to
        :param sel: Selection to write to file, None (default) to write
            all
        """

        # from iceprofret.txt in the Evans distribution:
        #
        # Six header lines:
        #    Number of pixels
        #    Number of channels, total number of elements
        #    Channel IDs or names (matches channel IDs in retrieval database)
        #    number of elements in each channel
        #    additive uncertainties for each element
        #    multiplicative uncertainties for each element
        # For each line:
        #    Time(hours)  Cosine viewing zenith angle   Measurement for each element


        if sel is None:
            sel = numpy.arange(self.data.size)

        with open(fn, 'wt') as fp:
            # number of pixels
            fp.write("{}\n".format(sel.size))
            # no channels / no elements; both equal to size of BT
            fp.write("{} {}\n".format(self.data.dtype["BT"].shape[0],
                                      self.data.dtype["BT"].shape[0]))
            # names of channels
            fp.write("{}\n".format(self.instrument.channel_string(pre="im",
                                                                  width=True)))

            # no. element per channel
            fp.write(("1 " * len(self.instrument.channels)).strip())
            fp.write("\n")

            # additive uncertainties
            for chan in self.instrument.channels:
                fp.write(str(chan.noise) + " ")
            fp.write("\n")

            # multiplicative uncertainties (not implemented)
            for chan in self.instrument.channels:
                fp.write("0 ")
            fp.write("\n")

            for elem in sel:
                # time; n/a, so write fake
                fp.write("0.0 ")
                # cos(angle); only nadir implented
                fp.write("1.0 ")
                # write channel BT's
                for chan_bt in self.data[elem]["BT"]:
                    fp.write(str(chan_bt) + " ")
                fp.write("\n")


    def write_mat(self, fn, fields=None, sel=None):
        """Write to MAT file.

        Use :func:`scipy.io.savemat` to write the database to a
        MATLAB-style .mat file.
        
        :param fn: Filename to write to
        :param fields: List of fieldnames to write.  Must be subset of
            ``self.data.dtype.names``.  Defaults to ``None``, which means
            to write all fields.
        :param sel: Indices to write.  Defaults to ``None``, which means
            all scenes.
        """
        if fields is None:
            fields = list(self.data.dtype.names)
        if sel is None:
            sel = slice(None)
        print(now(), "Writing to {}".format(fn))
        scipy.io.savemat(fn, dict(data=self.data[fields][sel],
                                  ice_axis=self.ice_axis),
                         appendmat=False,
                         do_compression=True, oned_as="column")

class LookupTable(abc.ABC):
    """Use a lookup table to consider similar measurements

    This table is used to represent a large set of measurements by a small
    set.  It takes as input a particular measurement, and returns as
    output a canonical measurements.  A measurement is n-dimensional,
    consisting of lat, lon, time of day, day of year, partial column,
    degrees of freedom.

    A use case is when we have a large set of measurements, but only error
    estimates for a subset of those.

    Implemented using a lookup table based on stats.bin_nd.  Bins are
    based on training data.  If newly presented data does not look like
    any pre-trained data, an error is raised.

    Binning based on PCAs is currently being implemented.

    Attributes::

        axdata.  Dictionary with keys corresponding to the axes to be
            considered in the lookup table; names should correspond to
            fields in data.  Each value is itself a dictionary with keys::

                nsteps: Number of steps in linspace

        bins
        db

    To create an instance, use either .fromData or .fromFile (if
    available).

    """

    #_loaded = False
    axdata = bins = db = None

    use_pca = False


    def compact_summary(self):
        """Return string with compact summary

        Suitable in filename
        """

        if self.use_pca:
            s = "PCA_{:s}_{:d}_{:.1f}".format(
                ",".join(self.axdata["PCA"]["fields"]),
                self.axdata["PCA"]["npc"],
                self.axdata["PCA"]["scale"])
        else:
            s = "-".join(
                    ["{:s}_{:d}".format(k,v["nsteps"])
                        for (k, v) in sorted(self.axdata.items())])
        return s

    def __repr__(self):
        return "<{}:{}>".format(self.__class__.__name__,
            self.compact_summary())


    @property
    def fields(self):
        if self.use_pca:
            return self.axdata["PCA"]["fields"]
        else:
            return self.axdata["fields"]

    def get_index_tuple(self, dat, full=False):
        """Get a tuple of indices for use in the lookup table

        Returns either the full tuple, or only the elements
        considering to the number of PCs considered originally.
        """
        if self.use_pca:
            fields = self.axdata["PCA"]["fields"]
            t = tuple(
                stats.bin_nd_sparse(self.pca.project(numpy.vstack(
                                        [dat[ax] for ax in fields]).T),
                                    self.bins).squeeze().tolist())
            if not full:
                t = t[:self.axdata["PCA"]["npc"]]

            return t
        else:
            fields = list(self.axdata.keys())
            return tuple(stats.bin_nd_sparse(
                    numpy.atleast_2d([dat[ax]
                    for ax in fields]), self.bins).squeeze().tolist())

    def get_index_tuples(self, data):
        """Yield tuples of indices for use in the lookup table
        """
        # FIXME: can be faster when `dat` is large
        for dat in data:
            yield (self.get_index_tuple(dat), dat)

       
    def lookup_all(self, data):
        # FIXME: can be faster when dat is large
        logging.info("Looking up {:d} radiances".format(data.size))
        bar = progressbar.ProgressBar(maxval=data.size,
            widgets=[progressbar.Bar("=", "[", "]"), " ",
                     progressbar.Percentage()])
        bar.start()
        for (i, dat) in enumerate(data):
            try:
                yield self.lookup(dat)
            except KeyError:
                logging.error("Not found for no. {:d}. :( "
                    "Should implement lookaround, enlarge LUT, or make it denser!".format(i))
                continue
                #yield None
        #yield from (self.lookup(dat) for dat in data)
            bar.update(i)
        bar.finish()

    def lookaround(self, dat):
        """Yield all neighbouring datapoints

        Look at all neighbouring datapoints.  NB those may be 2^N where N is
        the length of tup!  Very slow!
        """
        tup = self.get_index_tuple(dat)
        manytup = itertools.product(*[range(i-1,i+2) for i in tup])
        yield from (t for t in manytup if t in self.db)

    def _get_bins(self, data, axdata, pca=False):
        bins = []
        if pca:
            # This means axdata has a single key "PCA" with fields
            # “scale”.  It also means that `data` is in PCA space, i.e.
            # pca.Y.
#            rmin = numpy.nanmin(data, 0)
#            rmin -= 0.001*abs(rmin)
#            rmax = numpy.nanmax(data, 0)
#            rmax += 0.001*abs(rmax)
            # number of bins per PC
            nbins = numpy.ceil(self.pca.fracs*100*axdata["PCA"]["scale"])
            bins = [numpy.linspace(data[:, i].min(),
                                   data[:, i].max(),
                                   max(p, 2))
                        for (i, p) in enumerate(nbins)]
            return bins[:axdata["PCA"]["npc"]]
#            b = [self._get_bins_from_range(rmi, rma, axdata, "PCA")
#                    for (rmi, rma) in zip(rmin, rmax)]
#            raise NotImplementedError("Not implemented yet!")
        else:
            for ax in axdata.keys():
                if "range" in axdata[ax].keys():
                    (rmin, rmax) = axdata[ax]["range"]
                else:
                    rmin = numpy.nanmin(data[ax])
                    rmin -= 0.001*abs(rmin)
                    rmax = numpy.nanmin(data[ax])
                    rmax += 0.001*abs(rmax)

                b = self._get_bins_from_range(rmin, rmax, axdata, ax)
#                for case in tools.switch(axdata[ax].get("mode", "linear")):
#                    if case("linear"):
#                        b = numpy.linspace(rmin, rmax, axdata[ax]["nsteps"])
#                        break
#                    if case("optimal"):
#                        inrange = (data[ax] >= rmin) & (data[ax] <= rmax)
#                        b = scipy.stats.scoreatpercentile(data[ax][inrange],
#                                numpy.linspace(0, 100, axdata[ax]["nsteps"]))
#                        break
#                    if case():
#                        raise ValueError("ax {!s} unknown mode {!s}, I know "
#                            "'linear' and 'optimal'".format(axdata[ax], axdata[ax]["mode"]))
                bins.append(b)
            return bins
            # end for
        # end if

    @staticmethod
    def _get_bins_from_range(rmin, rmax, axdata, ax):
        """Small helper for _get_bins.

        From extrema and `axdata` description, get either linearly or
        logarithmically spaced bins.
        """
        for case in tools.switch(axdata[ax].get("mode", "linear")):
            if case("linear"):
                b = numpy.linspace(rmin, rmax, axdata[ax]["nsteps"])
                break
            if case("optimal"):
                inrange = (data[ax] >= rmin) & (data[ax] <= rmax)
                b = scipy.stats.scoreatpercentile(data[ax][inrange],
                        numpy.linspace(0, 100, axdata[ax]["nsteps"]))
                break
            if case():
                raise ValueError("ax {!s} unknown mode {!s}, I know "
                    "'linear' and 'optimal'".format(ax, axdata[ax]["mode"]))
        return b

    @staticmethod
    def _make_pca(data, axdata):
        
        fields = axdata["PCA"]["fields"]
        valid_range = axdata["PCA"]["valid_range"]
        if not all([issubclass(data[x].dtype.type, numpy.floating)
                        for x in fields]):
            logging.warning("Casting all data to float64 for PCA")

        data_mat = numpy.vstack([data[x] for x in fields]).T
        valid = numpy.all((data_mat > valid_range[0]) &
                          (data_mat < valid_range[1]), 1)
        return matplotlib.mlab.PCA(data_mat[valid, :])

    def lookup(self, dat):
        tup = self.get_index_tuple(dat)
        return self[tup]

    @staticmethod
    @abc.abstractmethod
    def _choose():
        ...

    @classmethod
    @abc.abstractmethod
    def fromData(cls):
        ...

    @abc.abstractmethod
    def __getitem__(self):
        ...

    @abc.abstractmethod
    def __setitem__(self):
        ...

class SmallLookupTable(LookupTable):
    """Lookup table small enough to be in memory
    """
    @classmethod
    def fromFile(cls, file):
        with open(file, 'rb') as fp:
            (axdata, bins, db) = pickle.load(fp)
        self = cls()
        self.axdata = axdata
        self.bins = bins
        self.db = db
        #self._loaded = True
        return self

    def propose_filename(self):
        return "similarity_db_{}".format(self.compact_summary())

    def toFile(self, file):
        """Store lookup table to a file
        """
        with open(file, 'wb') as fp:
            pickle.dump((self.axdata, self.bins, self.db), fp,
                    protocol=4)

    @classmethod
    def fromData(cls, data, axdata):
        """Build lookup table from data

        ``data`` should be a structured ndarray with dtype fields

        ``axdata`` should be a ``collections.OrderedDict`` where the keys
        refer to fields from `data` to use, and the values are
        dictionaries with the keys.  If regular binning is used (i.e. no
        PCA), those keys are:

            nsteps (mandatory)
                number of steps in binning data
            mode
                string that can be either "linear" (use linspace between
                extremes) or "optimal" (choose bins based on percentiles so
                1-D binning would create equal content in each).
                Defaults to linear.
            range
                tuple with (min, max) of range within which to bin data.
                Defaults to extremes of data.

        
        It is also possible to bin based on PCA.  In this case, ``axdata''
        should have a single key "PCA".  When binning based on PCA, the
        number of bins per PC are proportional the the proportion of
        variance along each PC axis (pca.fracs).  By default, the number
        of bins is the % of variability associated with the axis, i.e. if
        the first PC explains 67% of variability and the second 25%, there
        will be 67 and 25 bins, respectively.  This can be scaled by
        setting the key `scale` to something other than one.

            fields
                Sequence of strings: what fields to use in PCA-based
                analysis.

            npc
                Integer, how many PCs to consider

            scale
                Float, defaults to 1.0, for scaling the number of bins.

            valid_range
                Before performing PCA, require that ALL input vectors are
                within this range, otherwise discard.

        """
        self = cls()
        self.use_pca = list(axdata.keys()) == ["PCA"]
        self.pca = None
        if self.use_pca:
            # _make_pca considers axdata["PCA"]["fields"]
            self.pca = self._make_pca(data, axdata)
            # _get_bins considers axdata["PCA"]["npc"]
            #                 and axdata["PCA"]["scale"]
            bins = self._get_bins(self.pca.Y, axdata, pca=True)
            binned_indices = stats.bin_nd(
                [self.pca.Y[:, i] for i in range(self.pca.Y.shape[1])])
        else:
            fields = axdata.keys()
            bins = self._get_bins(data, axdata, pca=False)
            binned_indices = stats.bin_nd(
                [data[ax] for ax in fields], bins)
        db = {}
        # Do something for every bin.  `_choose` is implemented in another
        # class, it might do a count, it might choose one, or it might
        # choose all.
        for ii in itertools.product(*(range(i) for i in
                binned_indices.shape)):
            # ii should be a tuple that can be passed directly
            if binned_indices[ii].size > 0:
                db[ii] = data[self._choose(binned_indices[ii])]

        self.axdata = axdata
        self.bins = bins
        self.db = db
        #self._loaded = True
        return self

    def __getitem__(self, tup):
        return self.db[tup]

    def __setitem__(self, tup, val):
        self.db[tup] = val

    def keys(self):
        yield from self.db.keys()
 
class LargeLookupTable(LookupTable):
    """Lookup table too large in memory, mapped to directory

    """
    basename = "bucket/{coor:s}/contents.npy.xz"
    _db = {}
    _maxcache = 1e9 # 100 MB
    _N = 0

    def propose_dirname(self):
        return "large_similarity_db_{}".format(self.compact_summary())

    @classmethod
    def fromData(cls, data, axdata, use_pca=False):
        # docstring copied from SmallLookupTable

        self = cls()
        self.use_pca = use_pca
        if use_pca:
            self.pca = self._make_pca(data, axdata)
            bins = self._get_bins(self.pca.Y, axdata, pca=True)
        else:
            bins = self._get_bins(data, axdata, pca=False)
        self.axdata = axdata
        self.bins = bins
        if not self.bucket_dir().is_dir():
            self.bucket_dir().mkdir(parents=True)
        self.storemeta()
        self.addData(data)
        return self
    fromData.__doc__ = SmallLookupTable.__doc__

    def addData(self, data):
        """Add a lot of data
        """
        k = set()
        bar = progressbar.ProgressBar(maxval=len(data),
            widgets=[progressbar.Bar("=", "[", "]"), " ",
                    progressbar.Percentage()])
        bar.start()
        for (i, (t, contents)) in enumerate(self.get_index_tuples(data)):
            if t in k:
                cur = self[t]
#                contents = contents[numpy.array([contents[i] not in cur
#                        for i in range(contents.size)])]
#                if contents.size > 0:
                if contents not in cur:
                    self[t] = numpy.hstack((self[t], numpy.atleast_1d(contents)))
            else:
                self[t] = numpy.atleast_1d(contents)
            k.add(t)
            bar.update(i+1)
        bar.finish()
        self.dumpcache() # store and clear
        self.storemeta()

    @classmethod
    def fromDir(cls, arg):
        """Initialise from directory

        `arg` can be either a directory (str or pathlib.Path), or a
        dictionary of axdata describing such (see fromData docstring).
        """
        self = cls()

        if isinstance(arg, dict):
            self.axdata = arg
            self.use_pca = "PCA" in arg
            dir = self.bucket_dir()
        else:
            dir = arg

        with (pathlib.Path(dir) / "info.npy").open(mode="rb") as fp:
            logging.info("Reading into {!s}".format(pathlib.Path(dir)))
            (self.axdata, self.bins) = pickle.load(fp)
        if "PCA" in self.axdata:
            self.use_pca = True
            with (pathlib.Path(dir) / "pca.npy").open(mode="rb") as fp:
                self.pca = pickle.load(fp)
        return self

    def keys(self):
        """Yield keys one by one.  Reads from directory, no caching!
        """
        for p in self.bucket_dir().iterdir():
            if p.name.startswith("bucket_") and p.name.endswith(".npy.xz"):
                yield tuple(int(s) for s in p.name[7:-7].split("-"))

    def __setitem__(self, tup, data):
        self._db[tup] = data
        if len(self._db) > self._N: # recalculate size
            totsize = sum(v.nbytes for v in self._db.values())
            if totsize > self._maxcache:
                logging.debug("Size {:,} exceeds max cache {:,}, "
                              "dumping {:d} keys".format(totsize,
                              self._maxcache, len(self._db)))
                self.dumpcache()
            else:
                self._N += 10 # i.e. after every 10 new entries, check size

    def __getitem__(self, tup):
        if tup in self._db: # cached
            return self._db[tup]
        else:
            path = self.bucket_name(tup)
            if not path.exists():
                raise KeyError("No entry for {!s}".format(tup))
            with lzma.open(str(path), mode="rb") as fp:
                try:
                    v = numpy.load(fp)
                except Exception as e:
                    raise type(e)(str(e) + " while reading {!s}".format(
                        path)).with_traceback(sys.exc_info()[2])
            self._db[tup] = v
            return v

    def dumpcache(self):
        sizes = [v.size for v in self._db.values()]
        logging.info("Dumping cache for {:,} profiles in {:d} buckets to {!s}".format(
                    sum(sizes), len(self._db), self.bucket_dir()))
        bar = progressbar.ProgressBar(maxval=len(self._db), 
                    widgets=[progressbar.Bar('=', '[', ']'), ' ',
                    progressbar.Percentage()])
        bar.start()
        newdirs = 0
        counts = numpy.zeros(dtype=numpy.uint32, shape=(max(sizes)+1),)
        for (i, (k, v)) in enumerate(self._db.items()):
            path = self.bucket_name(k)
            if not path.parent.is_dir():
                #logging.info("Creating directory {!s}".format(path.parent))
                path.parent.mkdir(parents=True)
                newdirs += 1
#            if v.nbytes > 1e6:
#                logging.debug("Storing {:d}/{:d}, {:,} bytes to {!s}".format(
#                    i, len(self._db), v.nbytes, path))
            counts[v.size] += 1
            with lzma.open(str(path), mode="wb") as fp:
                numpy.save(fp, v)
            bar.update(i+1)
        bar.finish()
        logging.info("Stored cache.  Created {:d} new directories. "
                     "Profiles per bucket histogram: {!s}".format(newdirs, counts))

        self.clearcache()

    def storemeta(self):
        """Store metadata for database
        """
        d = self.bucket_dir()
        with (d / "info.npy").open(mode="wb") as fp:
            pickle.dump((self.axdata, self.bins), fp, protocol=4)
        if self.use_pca:
            with (d / "pca.npy").open(mode="wb") as fp:
                pickle.dump(self.pca, fp, protocol=4)

    def loadmeta(self):
        """Load metadata for database
        """
        d = self.bucket_dir()
        with (d / "info.npy").open(mode="rb") as fp:
            (self.axdata, self.bins) = pickle.load(fp)

    def clearcache(self):
        self._db.clear()
        self._N = 0

    def bucket_dir(self):
        return (pathlib.Path(config.conf["main"]["lookup_table_dir"]) /
                self.propose_dirname())

    def bucket_name(self, coor):
        """Return full path to bucket at coor
        """
        return (self.bucket_dir() /
                self.basename.format(coor="/".join("{:02d}".format(x) for x in coor)))

class SimilarityLookupTable(LookupTable):
    def propose_filename(self):
        return "tanso_similarity_db_{}".format(self.compact_summary())

    @staticmethod
    def _choose(data):
        """Choose one of the data to use for building the db
        """

        return random.choice(data)


class FullLookupTable(LookupTable):
    """Like a similarity lookup table, but keeps all entries

    """

    @staticmethod
    def _choose(data):
        return data


class CountingLookupTable(LookupTable):
    """Provide counting only, effectively creating a histogram.
    """

    @staticmethod
    def _choose(data):
        return data.size

class SmallSimilarityLookupTable(SmallLookupTable, SimilarityLookupTable):
    pass

class LargeSimilarityLookupTable(LargeLookupTable, SimilarityLookupTable):
    pass

class SmallFullLookupTable(SmallLookupTable, FullLookupTable):
    pass

class LargeFullLookupTable(LargeLookupTable, FullLookupTable):
    pass

class SmallCountingLookupTable(SmallLookupTable, CountingLookupTable):
    pass



"""Various time utilities
"""

import datetime

def mean_local_solar_time(utctime, lon):
    """Calculate mean local solar time.

    Calculates the mean local solar time for a specific longitude.
    This is not the true local solar time because it does not take into
    account the equation of time.  It is purely based on the hour angle of
    the Sun.  Do not use for astronomical calculations!

    :param time utctime: Time in UTC.  Should be a datetime.time object.
    :param float lon: Longitude in degrees.  Can be either in [-180, 180]
        or in [0, 360].
    :returns time: Time in mean local solar time.
    """

    hours_offset = lon/15
    dummy_datetime = datetime.datetime.combine(datetime.date.today(), utctime)
    new_dummy = dummy_datetime + datetime.timedelta(hours=hours_offset)
    return new_dummy.time()


def dt_to_doy_mlst(dt, lon):
    """From a datetime object, get day of year and MLST

    From a datetime object, calculate the day of year and the mean local
    solar time.

    :param datetime.datetime dt: Datetime object
    :param float lon: Longitude (needed for mean local solar time)
    :returns: (doy, mlst) where `doy` is a float between 0 and 366,
        and `mlst` is a float between 0 and 24.
    """

    utctime = dt.time()
    mlst = mean_local_solar_time(utctime, lon)
    frac_mlst = mlst.hour + mlst.minute/60 + mlst.second/3600

    dd = dt.date()
    doy = (dd - dd.replace(month=1, day=1)).days + 1
    return (doy, frac_mlst)



#!/usr/bin/env python

# coding: utf-8

"""Various small geo functions

"""

def shift_longitudes(lon, rng):
    """Shift longitudes to [-180, 180] or [0, 360]

    No longitudes should exist outside [-180, 360] in the input data.

    :param lon: Longitudes to be shifted
    :param rng: Either (-180, 180) or (0, 360)
    :returns: New longitudes
    """

    if lon.ptp() >= 360:
        raise ValueError(("Invalid longitude range ({}-{}) in "
            "input").format(lon.min(), lon.max()))
    if rng == (-180, 180):
        lon[lon>180] -= (360 * (lon.max()//360 + 1))
        lon[lon<-180] += (360 * abs(lon.min()//360 + 1))
    elif rng == (0, 360):
        raise NotImplementedError("Not implemented")
        #lon[lon<0] += 360
    else:
        raise ValueError("Unknown range: {}".format(rng))

    return lon

def valid_geo(arr):
    """Check if geo-data are valid.

    Should be contains in arr["lat"], arr["lon"].

    Valid is -90 <= lat <= 90, -180 <= lon <= 180
    """

    return (arr["lat"].min() >= -90 and
            arr["lat"].max() <= 90 and
            arr["lon"].min() >= -180 and
            arr["lon"].min() <= 180)



"""Ground-based remote sensing datasets
"""

import string
import datetime
import ast
import itertools
import logging

import numpy

import pytz
try:
    import pyhdf.SD
except ImportError:
    logging.error("Failed to import pyhdf, some readers will fail")

from .. import dataset
from .. import geo
from .. import io

class NDACCAmes(dataset.MultiFileDataset):
    """NDACC Ames-style file

    Documented at http://www.ndsc.ncep.noaa.gov/data/formats/
    """

    re = r"eutc(?P<year>\d{2})(?P<month>\d{2})\.sgf"
    name = "PEARL Bruker IFS"

    type_core = [(spec + "_" + tp, numpy.uint16 if tp=="n" else numpy.float32)
        for spec in ("O3", "HCL", "HF", "HNO3", "CLONO2", "N2O", "CO", "CH4")
        for tp in ("total", "total_e", "ss", "ss_e") +
            (("ts", "ts_e") if spec in {"O3", "N2O", "CO", "CH4"} else ()) +
            ("n",)]
    dtype = (
        [("time_yearfrac", numpy.float32)] +
        [("doy", numpy.uint16), ("year", numpy.uint16), 
         ("month", numpy.uint8), ("day", numpy.uint8),
         ("lat", numpy.float32), ("lon", numpy.float32),
         ("elev", numpy.float32)] +
        type_core)

    def get_time_from_granule_contents(self, path):
        # Read the entire granule, as the end information is sometimes
        # incorrect
        M = self.read(path)
        return (min(M["time"]).item(), max(M["time"]).item())
#        with open(path, 'rt', encoding="ascii") as fp:
#            for _ in range(7):
#                fp.readline()
#            y1, m1, d1, y2, m2, d2 = tuple(
#                int(d) for d in fp.readline().split())
#        return (datetime.datetime(y1, m1, d1, 0, 0, 0),
#                datetime.datetime(y2, m2, d2, 23, 59, 59))
             
    def _read(self, path, fields="all"):
        """Read Bruker data in NDACC Gaines format

        Returns a masked array
        """

        with open(path, 'rt', encoding='ascii') as fp:
            # first 10 lines not relevant for now
            header = ''.join(fp.readline() for _ in range(10))

            # no. of measurements per record
            nv = int(fp.readline().strip())

            # factors for each record
            vscal = io.collect_values(fp, nv, numpy.float32)

            # fillers for each record
            vmiss = io.collect_values(fp, nv, numpy.float32)

            # next N=48 lines contain variable names
            varnames = [fp.readline().strip() for _ in range(nv)]

            # the same for aux variables
            nauxv = int(fp.readline().strip())
            ascal = io.collect_values(fp, nauxv, numpy.float32)
            amiss = io.collect_values(fp, nauxv, numpy.float32)
            varnames_aux = [fp.readline().strip() for _ in range(nauxv)]

            # special comments
            nscoml = int(fp.readline().strip())
            scom = ''.join(fp.readline() for _ in range(nscoml))

            # normal comments
            nncoml = int(fp.readline().strip())
            ncom = ''.join(fp.readline() for _ in range(nncoml))

            # and now the data!
            # ...which needs a prettier dtype
            L = []
            while True:
                try:
                    v = io.collect_values(fp, 1+nauxv+nv, self.dtype)
                except EOFError:
                    break
                else:
                    dt = numpy.datetime64(datetime.date(v["year"],
                        v["month"], v["day"]), 'D').astype("<M8[us]")
                    new = numpy.empty(dtype=[("time",
                        dt.dtype)]+v.dtype.descr, shape=())
                    new["time"] = dt
                    for field in v.dtype.names:
                        new[field] = v[field]
                    L.append(new if fields=="all" else new[fields])

        # now I have an array with fractional years as the time-axis.
        # I want to have datetime64
        A = numpy.array(L)
        #dts = [datetime.datetime(numpy.floor(d), 1, 1, 0, 0, 0) +
        #       datetime.timedelta(days=(365+calendar.isleap(2010))*(d-numpy.floor(d)))
        #       for d in A["time"]]
        #dtn = dict(A.dtype.fields)
        #dtn["time"] = (numpy.dtype("datetime64[us]"), 0)
        #AA = numpy.empty(dtype=dtn, shape=A.size)
        #AA["time"] = dts
        #for fld in set(A.dtype.names) - {'time'}:
        #    AA[fld] = A[fld]
        #return AA

        # apply masks and factors
        # No, rather not.  MaskedArrays are buggy.
#        M = A.view(numpy.ma.MaskedArray)
        M = A
        for (i, field) in enumerate(self.type_core):
#            M.mask[field[0]] = A[field[0]] == vmiss[i]
            A[field[0]] *= vscal[i]

        # lats are secretly wrongly typed
        M["lat"] /= 100
        M["lon"] /= 100

        # and lons want to be in -180, 180
        M["lon"] = geo.shift_longitudes(M["lon"], (-180, 180))
        return M


class Eureka_PRL_CH4_HDF(dataset.MultiFileDataset, dataset.ProfileDataset):
    # NOTE: there is a bug in older versions of Python-hdf4 that causes it to
    # crash on some HDF files.  The Eureka Bruker CH4 HDF files happen to
    # have a crash on
    # CH4.MIXING.RATIO.VOLUME_ABSORPTION.SOLAR_UNCERTAINTY.SYSTEMATIC.COVARIANCE
    # The bug was fixed 2014-11-26

    # For PEARL collocating with TANSO, 80% of profiles have >50%
    # sensitivity between 3.9 and 29.9 km, and 50% have >50% sensitivity
    # between 3.9 km and 31.5 km
    # For PEARL collocation with ACE, 80% of profiles have >50%
    # sensitivity between 3.9 km and 26.9 km, and 50% have >50%
    # sensitivity between 0.8 km and 29.9 km

    A_needs_converting = False
    A_needs_swapping = True

    range = (3.5e3, 30e3)

    # It does appear timezones are now fixed
    #timezone = "Etc/GMT-5"
    timezone = "UTC"

    # specific field for Eureka PEARL HDF
    altitude_boundaries = None

    aliases = {"CH4_profile": "CH4_VMR",
        "S_CH4_profile": "CH4_SA_random",
        "ap": "CH4_ap",
        "ak": "CH4_ak"}

    _nlev = 47
    _dtp = [("time", "datetime64[s]"),
            ("lat", "f4"),
            ("lon", "f4"),
            ("p0", "f4"),
            ("z0", "f4"),
            ("T0", "f4"),
            ("z", "f4", _nlev),
            ("p", "f4", _nlev),
            ("T", "f4", _nlev),
            ("CH4_VMR", "f4", _nlev),
            ("CH4_ak", "f4", (_nlev, _nlev)),
            ("CH4_ap", "f4", _nlev),
            ("CH4_SA_random", "f4", (_nlev, _nlev)),
            ("CH4_SA_system", "f4", (_nlev, _nlev)),
            ("CH4_pc", "f4", _nlev),
            ("CH4_ap_pc", "f4", _nlev),
            ("CH4_tc", "f4"),
            ("CH4_ap_tc", "f4"),
            ("CH4_ak_tc", "f4", _nlev),
            ("delta_CH4_tc_random", "f4"),
            ("delta_CH4_tc_system", "f4"),
            ("sza", "f4"),
            ("saa", "f4"),
            ("H2O_VMR", "f4", _nlev)
            ]

    _trans = {"DATETIME": "time",
        "LATITUDE.INSTRUMENT": "lat",
        "LONGITUDE.INSTRUMENT": "lon",
        "ALTITUDE.INSTRUMENT": "z0",
        "SURFACE.PRESSURE_INDEPENDENT": "p0",
        "SURFACE.TEMPERATURE_INDEPENDENT": "T0",
        "ALTITUDE": "z",
        "PRESSURE_INDEPENDENT": "p",
        "TEMPERATURE_INDEPENDENT": "T",
        "CH4.MIXING.RATIO.VOLUME_ABSORPTION.SOLAR": "CH4_VMR",
        "CH4.MIXING.RATIO.VOLUME_ABSORPTION.SOLAR_APRIORI": "CH4_ap",
        "CH4.MIXING.RATIO.VOLUME_ABSORPTION.SOLAR_AVK": "CH4_ak",
        "CH4.MIXING.RATIO.VOLUME_ABSORPTION.SOLAR_UNCERTAINTY.RANDOM.COVARIANCE":
            "CH4_SA_random",
        "CH4.MIXING.RATIO.VOLUME_ABSORPTION.SOLAR_UNCERTAINTY.SYSTEMATIC.COVARIANCE":
           "CH4_SA_system",
        "CH4.COLUMN.PARTIAL_ABSORPTION.SOLAR": "CH4_pc",
        "CH4.COLUMN.PARTIAL_ABSORPTION.SOLAR_APRIORI": "CH4_ap_pc",
        "CH4.COLUMN_ABSORPTION.SOLAR": "CH4_tc",
        "CH4.COLUMN_ABSORPTION.SOLAR_APRIORI": "CH4_ap_tc",
        "CH4.COLUMN_ABSORPTION.SOLAR_AVK": "CH4_ak_tc",
        "CH4.COLUMN_ABSORPTION.SOLAR_UNCERTAINTY.RANDOM.STANDARD":
            "delta_CH4_tc_random",
        "CH4.COLUMN_ABSORPTION.SOLAR_UNCERTAINTY.SYSTEMATIC.STANDARD":
            "delta_CH4_tc_system",
        "ANGLE.SOLAR_ZENITH.ASTRONOMICAL": "sza",
        "ANGLE.SOLAR_AZIMUTH": "saa",
        "H2O.MIXING.RATIO.VOLUME_ABSORPTION.SOLAR": "H2O_VMR"}
    #_invtrans = {v: k for k, v in _trans.items()}
    _fld_copy_scal = {"lat", "lon", "p0", "T0", "z0", "CH4_tc",
        "CH4_ap_tc", "sza", "saa",
        "delta_CH4_tc_random", "delta_CH4_tc_system"}
    _fld_copy_onevec = {"z"}
    _fld_copy_vec = {"p", "T", "CH4_VMR", "CH4_pc", "CH4_ap_pc",
        "CH4_ak_tc", "H2O_VMR", "CH4_ap"}
    _fld_copy_mat = {"CH4_ak", "CH4_SA_system", "CH4_SA_random"}
    _fld_copy_not = {"time"} # specially treated
    def _read(self, path=None, fields="all"):
        """Read granule"""
        if path is None:
            path = self.srcfile

        sd = pyhdf.SD.SD(path)
        (n_ds, n_attr) = sd.info()

        n_elem = sd.select(0).info()[2]
        M = numpy.empty(shape=(n_elem,), dtype=self._dtp)
        dtm_mjd2k = sd.select(sd.nametoindex("DATETIME")).get()
        dtm_mjd2k_s = dtm_mjd2k * 86400
        M["time"] = (numpy.datetime64(datetime.datetime(2000, 1,
                                                         1, 0, 0, 0)) +
                         dtm_mjd2k_s.astype("timedelta64[s]"))
        # It appears timez are still in UTC-0500.  Correct accordingly.
        tz = pytz.timezone(self.timezone)
        M["time"] = [t + numpy.timedelta64(tz.utcoffset(t)) for t in M["time"]]
        # check direction as I may need to turnaround the data
        z = sd.select(sd.nametoindex("ALTITUDE")).get()
        direc = int(numpy.sign(z[-1]-z[0]))
        # simple copy
        for (full, short) in self._trans.items():
            sds = sd.select(sd.nametoindex(full))
            if short in self._fld_copy_scal:
                M[short] = sds.get()
            elif short in self._fld_copy_onevec:
                M[short] = sds.get()[::direc]
            elif short in self._fld_copy_vec:
                M[short] = sds.get()[:, ::direc]
            elif short in self._fld_copy_mat:
                M[short] = sds.get()[:, :, ::direc][:, ::direc, :]
            elif short in self._fld_copy_not:
                pass
            else:
                logging.error("Don't know where to put {} ({})!".format(full, short))

            (offset, factor, unit) = sds.attributes()["VAR_SI_CONVERSION"].split(';')
            factor = ast.literal_eval(factor)
            if not unit in {"rad", "mol m-2", "s"}:
                M[short] *= factor

        self.altitude_boundaries = sd.select(
            sd.nametoindex("ALTITUDE.BOUNDARIES")).get()

        # Now done above
        #M["p0"] *= HECTO
        #M["p"] *= HECTO
        #M["z"] *= KILO
        #M["z0"] *= KILO
        #M["CH4_VMR"] *= PPM

#        for i in range(n_ds):
#            (nm, rank, dims, tp, n_attr) = sd.select(i).info()
#            if (rank==1 and dims==n_elem):
#                dtp.append((nm, "<f4"))
#            elif (rank>1 and dims[0]==n_elem):
#                dtp.append((nm, "<f4", dims[1:]))

        return M



"""Read and write Arts XML types

This module contains classes for reading and writing Arts XML types.
It is built on top of xml.etree.ElementTree, available since Python 2.5,
but this module relies on the version 3.3 or newer.  Most users will only
need the 'parse' function.

FIXME:
- Does <arts> tag always contain exactly one child?
"""

import xml.etree.ElementTree

from . import _handlers


class ArtsElement(xml.etree.ElementTree.Element):
    """Element with value interpretation
    """
    def value(self):
        try:
            return getattr(_handlers, self.tag)(self)
        except AttributeError:
            raise ValueError("Don't know how to handle <{}>!".format(self.tag))

def parse(source):
    """Parse ArtsXML file from source.

    This function is very similar to xml.etree.ElementTree.parse, except
    that elements will be of type ArtsElement rather than
    xml.etree.Element.  See documentation for ArtsElement for how this
    helps you.
    """
    return xml.etree.ElementTree.parse(source,
            parser=xml.etree.ElementTree.XMLParser(
                target=xml.etree.ElementTree.TreeBuilder(
                    element_factory=ArtsElement)))




#!/usr/bin/env python

# coding: utf-8

"""Various small physics functions

Mostly obtained from PyARTS
"""

import logging
import numbers
import datetime
import calendar
import itertools

import numpy
import scipy.interpolate
import matplotlib
import matplotlib.dates

import numexpr
import pyproj
import pint


from .constants import (h, k, R_d, R_v, c, M_d, M_w, micro, R)
#from . import constants as c
from . import math as pamath
from . import time as pytime
from . import tools
from . import graphics
from . import stats
from . import config
from . import io as pyio
from .units import ureg

from typhon.physics.units.em import (FwmuMixin, SRF, planck_f,
    specrad_wavenumber2frequency, specrad_frequency_to_planck_bt)

class AKStats:

    filename = "sensitivity_{mode}_matrix_{name}."
    #cmap = "afmhot_r"
    #cmap = trollimage.colormap.spectral
    cmap = "Spectral_r"

    def __init__(self, aks, name="UNDEFINED"):
        self.aks = aks.copy()
        with numpy.errstate(invalid="ignore"):
            self.aks[self.aks<=-999] = numpy.nan
        self.name = name

    def summarise(self, data):
        """Look at various statistics.
        """

        with numpy.errstate(invalid="warn"):
            self.plot_sensitivity_range(z=data["z"])
            self.plot_sensitivity_density(z=data["z"])
        self.plot_histogram()
        self.summarise_dof_stats(data)

    def dofs(self):
        """Calculate degrees of freedom.

        According to Rodgers (2000), equation (2.80), page 37.

        This is the trace of the averaging kernels.
        """

        # Circumvent https://github.com/numpy/numpy/issues/5560
        #return self.aks.trace(axis1=1, axis2=2)
        # Is actually not effectively circumvented (anymore?)
        #return numpy.trace(self.aks, axis1=1, axis2=2)
        if isinstance(self.aks, numpy.ma.MaskedArray):
            return self.aks.data.trace(axis1=1, axis2=2)
        else:
            return self.aks.trace(axis1=1, axis2=2)

    def sensitivities(self):
        """Calculate sensitivities.

        According to:
          R. L. Batchelor et al.: Ground-based FTS comparisons ad ACE
          validation at Eureka during IPY.  Page 57.
        Sum of each row of the averaging kernel matrix defines sensitivity
        to measurument.  Note that I - A is the sensitivity to the a priori,
        soand the sum of the rows of (I - A) + the sum of the rows of A
        equals 1; so these two could be interpreted as percentages.
        """

        return self.aks.sum(1)

    def sensitivity_density_matrix(self, sens_fractions=numpy.linspace(0, 1, 11)):
        """What fraction of profiles have sensitivity >x at level y?

        :param ndarray sens_fractions: Fractions x to consider
        :returns: (sens_fractions, sens_mat),
            where sens_mat is a matrix with fraction at each level with at
            least sensitivity x.
        """

        sensitivities = self.sensitivities()
        with numpy.errstate(invalid="ignore"):
            sensmat = numpy.vstack([(sensitivities>x).sum(0) / sensitivities.shape[0]
                for x in sens_fractions])
        return (sens_fractions, sensmat)

    def sensitivity_range_matrix(self,
        sens_fractions = numpy.linspace(0, 1, 11),
        sens_counters = None):
        """
        For each profile, how many layers have sensitivity of at least x?
        
        Creates a "sensitivity score" matrix.  How many profiles have at
        least y layers of sensitivity >= x?

        :param ndarray sens_fractions:
            Array of shape (N,).
            Sensitivities to consider.  Defaults to 0, 0.1, ..., 1.0.
            Will be used to count no. of profiles with sensitivity larger
            than this fraction.
        :param ndarray sens_counters:
            Integer array of shape (p,),
            indicating the count as to how many profiles
            have at least this many layers with sensitivity larger than x.
            Defaults to arange(self.aks.shape[1]+1)
        :returns:
            Tuple (sens_fractions, sens_counters, sensmat), where sensmat
            is an N x p matrix, N being the number of sensitivities to
            consider, and p the counters.  In each coordinate define by
            the ararys sens_fractions and sens_counters, it contains the
            fraction of profiles that have at least p levels above
            sensitivity x.
        """
        
        if sens_counters is None:
            sens_counters = numpy.arange(self.aks.shape[1]+1)

        sensitivities = self.sensitivities()
        with numpy.errstate(invalid="ignore"):
            sensmat = numpy.vstack([((sensitivities >= x).sum(1)>=y).sum() 
                for x in sens_fractions
                for y in sens_counters]).reshape(
                    sens_fractions.shape[0], sens_counters.shape[0])
        return (sens_fractions, sens_counters, sensmat / self.aks.shape[0])

    def sensitivity_range_matrix_z(self, z,
            arr_dz = numpy.linspace(0, 30e3, 11),
            arr_sens = numpy.linspace(0, 1, 11)):
        """Like sensitivity_range_matrix but with elevation units.

        Returns a matrix with the fraction of elements where sensitivity
        exceeds 'x' for a range of elevations 'dz'.

        :param z: Matrix containing elevations.  Shape must match
            self.aks.shape[1:].
        :param arr_dz: Array of delta-z to consider.
        :param arr_sens: Array of sensitivities to consider.
        :returns: (arr_dz, arr_sens, mat)
        """

        mat = numpy.zeros(shape=(arr_dz.size, arr_sens.size))
        sensitivities = self.sensitivities()
        with numpy.errstate(invalid="ignore"):
            allmsk = [sensitivities > x for x in arr_sens]
        # find highest and lowest z for each column, corresponding to msk
        #
        # Note: in some cases there are 'gaps', i.e. sensitivity mask
        # looks like [True, True, False, False, False, True, True, True,
        # True, True, True, True, True, False, False, ...].  In this case
        # we take the 'False' along for now.  This may also yield more
        # than one 'last'.

        for (msk_i, msk) in enumerate(allmsk):
            # first True in each column
            mskno = msk.nonzero()
            if not msk.any(): # all mat 0
                continue
            (_, ii) = numpy.unique(mskno[0], return_index=True) 
            makes_sense = msk.any(1)
            firsts = numpy.zeros(shape=sensitivities.shape[0])
            firsts[makes_sense] = mskno[1][ii] # NB: goes wrong if 0 True values
            firsts[~makes_sense] = -1

            # last True in each column is element before!
            lasts = numpy.zeros(shape=sensitivities.shape[0])
            lasts[makes_sense] = numpy.hstack((mskno[1][ii[1:]-1], mskno[1][-1]))
            lasts[~makes_sense] = 0

            lower = numpy.array([z[i, firsts[i]] for i in range(firsts.shape[0])])
            upper = numpy.array([z[i, lasts[i]] for i in range(lasts.shape[0])])
            dz = upper - lower
            # by handling > as false, nans are counted as not in range.
            # There may be nans in z
            with numpy.errstate(invalid="ignore"):
                for (dz_k, dz_lim) in enumerate(arr_dz):
                    mat[dz_k, msk_i] = (dz >= dz_lim).sum() / sensitivities.shape[0]

        return (arr_dz, arr_sens, mat)


    def plot_sensitivity_density(self,
            nstep=11,
            z=None):
        """Visualise where sensors are typically sensitive
        """

        (sens_frac, sensmat) = self.sensitivity_density_matrix()

        # regrid sensitivity matrix for z
        if z.ndim > 1:
            if (z.min(0) == z.max(0)).all():
                z = z[0, :]
            else:
                newz = numpy.nanmean(z, 0)
                logging.info("Regridding sensitivity matrices")
                A_new = pamath.regrid_matrix(sensmat, z, newz)
                logging.info("Done")
                z = newz

        # write some diagnostics
        for (i, f) in enumerate(sens_frac):
            for p in (0.2, 0.5, 0.8):
                makes_sense = z[sensmat[i, :]>p]
                if makes_sense.any():
                    logging.info(("Altitude range with at least "
                        "{:.0%} >{:.0%} sensitive: {:.1f}--{:.1f} km").format(
                            p, f, makes_sense.min()/1e3,
                            makes_sense.max()/1e3))
                else:
                    logging.info(("Never more than {:.0%} with "
                        "sensitivity {:.0%} :(").format(
                        sensmat[i, :].max(), f))
            
        #f = matplotlib.pyplot.figure()
        (f, a) = matplotlib.pyplot.subplots() # = f.add_subplot(1, 1, 1)
        cs = a.contourf(sens_frac, z, sensmat.T,
            numpy.linspace(0, 1, nstep),
            cmap=self.cmap)
        #cs.clabel(colors="blue")
        cb = f.colorbar(cs)
        a.set_xlabel("Sensitivity")
        a.set_ylabel("Elevation [m]")
        cb.set_label("Fraction")
        a.set_title("Elevation sensitivity density {}".format(self.name))
        a.grid(which="major", color="white")
        
        graphics.print_or_show(
            f, False, self.filename.format(mode="density_z", name=self.name),
            data=numpy.vstack([(sens_frac[i], z[j], sensmat[i,j])
                               for i in range(sens_frac.size)
                               for j in range(z.size)]).reshape(
                                    sens_frac.size, z.size, 3))

    def plot_sensitivity_range(self,
            nstep=11,
            z=None):
        """Visualise vertical range of sensitivities
        """

        # Degrees of freedom according to:
        #   Rodgers (2000)
        #   Equation (2.80), Page 37

        dofs = self.dofs()

        sensitivities = self.sensitivities()

        max_sensitivities = sensitivities.max(1)

        (sens_fractions, sens_counters, sensmat) = self.sensitivity_range_matrix()

        f = matplotlib.pyplot.figure()
        a = f.add_subplot(1, 1, 1)
        cs = a.contourf(sens_fractions, sens_counters, sensmat.T,
            numpy.linspace(0, 1, nstep),
            cmap=self.cmap)
        #cs.clabel(colors="blue")
        cb = f.colorbar(cs)
        a.set_xlabel("Sensitivity")
        a.set_ylabel("No. of levels")
        cb.set_label("Fraction")
        a.set_title(("Fraction of profiles with at least N layers "
                     "sensitivity > x"))
        graphics.print_or_show(
            f, False, self.filename.format(mode="range_n", name=self.name))

        (arr_dz, arr_sens, mat_z) = self.sensitivity_range_matrix_z(z=z)

        f = matplotlib.pyplot.figure()
        a = f.add_subplot(1, 1, 1)
        cs = a.contourf(arr_sens, arr_dz, mat_z,
            numpy.linspace(0, 1, nstep),
            cmap=self.cmap)
        #cs.clabel(colors="blue")
        cb = f.colorbar(cs)
        a.set_xlabel("Sensitivity")
        a.set_ylabel("Delta z [m]")
        cb.set_label("Fraction")
        a.set_title(("Fraction of profiles with sensitivity > x "
                     "throughout a certain vertical range"))
        graphics.print_or_show(
            f, False, self.filename.format(mode="range_z", name=self.name))

    def plot_histogram(self):
        dofs = self.dofs()
        (f, a) = matplotlib.pyplot.subplots()
        (N, x, p) = a.hist(dofs[numpy.isfinite(dofs)], 20)
        a.set_xlabel("DOFs")
        a.set_ylabel("count")
        a.set_title("histogram DOF collocated {}".format(self.name))
        graphics.print_or_show(f, False,
            "hist_dof_{}.".format(self.name),
                data = dofs[numpy.isfinite(dofs)]) # let pgfplots do the hist

    _dof_binners = dict(
        doy = dict(
            label = "Day of year",
            bins = numpy.linspace(0, 366, 24),
            invert = False,
            timeax = False),
        mlst = dict(
            label = "Mean local solar time",
            bins = numpy.linspace(0, 24, 24),
            invert = False,
            timeax = False),
        time = dict(
            label = "Date",
            invert = False,
            timeax = True),
        lat = dict(
            label = "Latitude",
            invert = False,
            timeax = False),
        lon = dict(
            label = "Longitude",
            timeax = False,
            invert = True),
        parcol = dict(
            label = "Par. col. CH4",
            timeax = False,
            invert = False),
#        dof = dict(
#            label = "DOF",
#            timeax = False,
#            invert = False),
        )

    def summarise_dof_stats(self, data):
        """Make and plot some DOF summaries
        """

        dofs = self.dofs()

        # get day of year and mean local solar time
        
        # (NB: outer zip effectively unzips)
        (doy, mlst) = zip(*(pytime.dt_to_doy_mlst(
                                dt.astype(datetime.datetime),
                                lon)
                for (dt, lon) in zip(data["time"], data["lon"])))

        # Prepare matplotlib date axes
        ml = matplotlib.dates.DayLocator(bymonthday=[1, 15])
        datefmt = matplotlib.dates.DateFormatter("%m/%d")

        D = dict(doy={}, mlst={}, lat={}, lon={}, parcol={})
        D["doy"]["data"] = numpy.array(doy)
        D["mlst"]["data"] = numpy.array(mlst)
        D["lat"]["data"] = data["lat"]
        D["lon"]["data"] = data["lon"]
        D["parcol"]["data"] = data["parcol_CH4"]
#        D["dof"]["data"] = dofs

        for k in D.keys():
            if k in self._dof_binners and "bins" in self._dof_binners[k]:
                D[k]["bins"] = self._dof_binners[k]["bins"]
            else:
                D[k]["bins"] = numpy.linspace(
                    D[k]["data"].min()*0.99, D[k]["data"].max()*1.01, 10)

        binners = sorted(D.keys())
        binned_indices = stats.bin_nd(
                            [D[k]["data"] for k in binners],
                            [D[k]["bins"] for k in binners])
        # make "fake" date range where I will use only month and day,
        # so I can use date-based plotting
        D["time"] = dict(data=numpy.array([(datetime.date(2015, 1, 1)
                        + datetime.timedelta(days=int(d))).toordinal()
                        for d in D["doy"]["data"]]))
        D["time"]["bins"] = numpy.array([(datetime.date(2015, 1, 1)
                        + datetime.timedelta(days=int(d))).toordinal()
                        for d in D["doy"]["bins"]])
        # replace "doy" by "time"
        binners[binners.index("doy")] = "time"

        combis = sorted([tuple(x) for x in 
                    {frozenset(x)
                        for x in itertools.product(
                            range(binned_indices.ndim),
                            range(binned_indices.ndim))} if len(x)>1])
        # combis = [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3), ...]
        for (i1, i2) in combis:
            names = dict(x=binners[i1], y=binners[i2])

            merged = stats.binsnD_to_2d(binned_indices, i1, i2)
        
            stat = {}
            for (nm, func) in [
                    ("median", numpy.median),
                    ("mad", pamath.mad)]:
                stat[nm] = numpy.array(
                    [(func(dofs[merged.flat[k]])
                        if merged.flat[k].size>0
                        else numpy.nan)
                            for k in range(merged.size)]
                    ).reshape(merged.shape)
            stat["count"] = numpy.array(
                [merged.flat[k].size for k in range(merged.size)]
                    ).reshape(merged.shape)
            stat[""] = None # special value for scatter

            if (stat["count"]>1).sum() < 2:
                continue # all in one bin, don't plot
            #

            for (statname, val) in stat.items():

#                for mode in ("pcolor", "scatter"):
                (f, a) = matplotlib.pyplot.subplots()
#                    if mode == "pcolor":
                if statname == "":
                    pc = a.scatter(D[names["x"]]["data"],
                                   D[names["y"]]["data"], 
                                   c=dofs,
                                   s=50,
                                   cmap=self.cmap)
                else:
                    val_masked = numpy.ma.masked_array(val,
                        numpy.isnan(val))
                    pc = a.pcolor(D[names["x"]]["bins"],
                                  D[names["y"]]["bins"],
                                  val_masked.T,
                                  cmap=self.cmap)
#                    elif mode == "scatter":
                cb = f.colorbar(pc)

                #for (axname, axis, fmt) in [
                for axlt in "xy":
                    axname = names[axlt]
                    axis = getattr(a, "{}axis".format(axlt))
                    if self._dof_binners[axname]["timeax"]:
                        axis.set_major_locator(ml)
                        axis.set_major_formatter(datefmt)
                        getattr(f, "autofmt_{}date".format(axlt))()
                    if self._dof_binners[axname]["invert"]:
                        getattr(a, "invert_{}axis".format(axlt))()
                    getattr(a, "set_{}label".format(axlt))(
                        self._dof_binners[axname]["label"])

                cb.set_label("DOF " + statname)
                a.set_title("DOF " + self.name)
                graphics.print_or_show(f, False,
                    "DOF_{}_{}_{}_{}.".format(
                    self.name, names["x"], names["y"], statname),
                        data=(numpy.vstack((D[names["x"]]["data"],
                            D[names["y"]]["data"], dofs)).T
                                if (statname=="" and names["x"]=="time") else None))
                # write "by hand" because of datetime
                if names["x"] == "time" and statname=="":
                    with open("{:s}/{:s}".format(pyio.plotdatadir(),
                            "DOF_time_{:s}_{:s}.dat".format(
                                self.name, names['y'])), 'w') as f:
                        for i in range(dofs.shape[0]):
                            f.write("{:%Y-%m-%d} {:.3f} {:.3f}\n".format(
                                datetime.date.fromordinal(D["time"]["data"][i]),
                                D[names['y']]["data"][i], dofs[i]))
        
def mixingratio2density(mixingratio, p, T):
    """Converts mixing ratio (e.g. kg/kg) to density (kg/m^3) for dry air.

    Uses the ideal gas law and the effective molar mass of Earth air.

    :param mixingratio: Mixing ratio [1]
    :param p: Pressure [Pa]
    :param T: Temperature [K]
    :returns: Density [kg/m^3]
    """

    # from ideal gas law, mass of 1 m³ of air:
    #
    # ρ = p/(R*T)

    m_air = p/(R_d*T)
    return mixingratio * m_air

def mixingratio2rh(w, p, T):
    """For water on Earth, convert mixing-ratio to relative humidity

    :param w: water vapour mixing ratio [1]
    :param p: pressure [Pa]
    :param T: temperature [K]
    :returns: relative humidity [1]
    """

    eps = R_d/R_v # Wallace and Hobbs, 3.14
    e = w/(w+eps)*p # Wallace and Hobbs, 3.59
    e_s = vapour_P(T)
    return e/e_s # Wallace and Hobbs, 3.64

def rh2mixingratio(rh, p, T):
    """Convert relative humidity to water vapour mixing ratio

    Based on atmlabs h2o/thermodynomics/relhum_to_vmr.m.

    :param rh: Relative humidity [1]
    :param p: Pressure [Pa]
    :param T: Temperature [K]
    :returns: Water vapour mixing ratio [1]
    """

    return rh * vapour_P(T) / p

def specific2mixingratio(q):
    """Convert specific humidity [kg/kg] to volume mixing ratio
    """

    # Source: extract_arts_1.f90

    eps = R_d/R_v
    return q / ( q + eps*(1-q) )


def vapour_P(T):
    """Calculate saturation vapour pressure.

    Calculates the saturated vapour pressure (Pa)
    of water using the Hyland-Wexler eqns (ASHRAE Handbook).

    (Originally in PyARTS)
    
    :param T: Temperature [K]
    :returns: Vapour pressure [Pa]
    """
    
    A = -5.8002206e3
    B = 1.3914993
    C = -4.8640239e-2
    D = 4.1764768e-5
    E = -1.4452093e-8
    F = 6.5459673
    
    Pvs = numpy.exp(A/T + B + C*T + D*T**2 + E*T**3 + F*numpy.log(T))
    return Pvs

def specific2iwv(z, q):
    """Calculate integrated water vapour [kg/m^2] from z, q

    :param z: Height profile [m]
    :param q: specific humidity profile [kg/kg]
    :returns: Integrated water vapour [kg/m^2]
    """

    mixing_ratio = specific2mixingratio(q)
    return pamath.integrate_with_height(z, mixing_ratio)

def rh2iwv(z, rh, p, T):
    """Calculate integrated water vapour [kg/m^2] from z, rh

    :param z: Height profile [m]
    :param rh: Relative humidity profile [1]
    :param p: Pressure profile [Pa]
    :param T: Temperature profile [T]
    :returns: Integrated water vapour [kg/m^2]
    """
    mixing_ratio = rh2mixingratio(rh, p, T)
    return pamath.integrate_with_height(z, mixingratio2density(mixing_ratio, p, T))

def mixingratio2iwv(z, r, p, T):
    """Calculate integrated water vapour [kg/m^2] from z, r

    :param z: Height profile [m]
    :param r: mixing ratio profile [kg/kg]
    :param p: Pressure profile [Pa]
    :param T: Temperature profile [T]
    :returns: Integrated water vapour [kg/m^2]
    """

    return pamath.integrate_with_height(z, mixingratio2density(r, p, T))

def spectral_to_channel(f_L, L_f, f_srf, w_srf):
    """From a spectrum of radiances and a SRF, calculate channel radiance

    The spectral response function may not be specified on the same grid
    as the spectrum of radiances.  Therefore, this function interpolates
    the spectral response function onto the grid of the radiances.  This
    is less bad than the reverse, because a spectral response function
    tends to be more smooth than a spectrum.

    **Approximations:**

    * Interpolation of spectral response function onto frequency grid on
      which radiances are specified.

    FIXME BUG!  The conversion is incorrect and should be done in terms of
    spectral radiances, not brightness temperatures.  Consider discussion
    with Jon.  Need to read up on this.  FIXME BUG!

    :param ndarray f_L: Frequencies for spectral radiances [Hz]
    :param ndarray L_f: Spectral radiances [various].  Can be in
        radiance units or brightness temperatures.  Innermost dimension
        must correspond to frequencies.
    :param ndarray f_srf: Frequencys for spectral response function [Hz]
    :param ndarray w_srf: Weights for spectral response function []
    :returns: Channel radxiance (same unit as L_f).
    """
    # Interpolate onto common frequency grid.  The spectral response
    # function is more smooth so less harmed by interpolation, so I
    # interpolate the SRF.
    f = scipy.interpolate.interp1d(f_srf, w_srf, bounds_error=False, fill_value=numpy.nan)
    w_on_L_grid = f(f_L)
    #ch_BT = (w_on_L_grid * L_f).sum(-1) / (w_on_L_grid.sum())
    # due to numexpr limitation, do sum seperately
    ch_BT_tot = numexpr.evaluate("sum(w_on_L_grid * L_f, {:d})".format(
                                L_f.ndim-1))
    ch_BT = ch_BT_tot / w_on_L_grid.sum()
    return ch_BT


def vmr2nd(vmr, T, p):
    """Convert volume mixing ratio [] to number density

    :param vmr: Volume mixing ratio or volume fraction.  For example,
        taking methane density in ppmv, first multiply by `constants.ppm`,
        then pass here.
    :param T: Temperature [K]
    :param p: Pressure [Pa]
    :returns: Number density in molecules per m^3
    """

    # ideal gas law: p = n_0 * k * T
    return  vmr * p / (k * T)

def p2z_oversimplified(p):
    """Convert pressure to altitude with oversimplified assumptions.

    Neglects the virtual temperature correction, assumes isothermal
    atmosphere with pressure dropping factor 10 for each 16 km.  Use a
    better function...

    :param p: Pressure [Pa]
    :returns: Altitude [m]
    """

    return 16e3 * (5 - numpy.log10(p) )

@tools.validator
def p2z_hydrostatic(p:numpy.ndarray,
    T:numpy.ndarray,
    h2o,
    p0:(numpy.number, numbers.Number, numpy.ndarray),
    z0:(numpy.number, numbers.Number, numpy.ndarray),
    lat:(numpy.number, numbers.Number, numpy.ndarray)=45,
    z_acc:(numpy.number, numbers.Number, numpy.ndarray)=-1,
    ellps="WGS84",
    extend=False):
    """Calculate hydrostatic elevation

    Translated from
    https://www.sat.ltu.se/trac/rt/browser/atmlab/trunk/geophysics/pt2z.m

    WARNING: seems to get siginificant errors.  Testing with an ACE
    profile between 8.5 and 150 km, I get errors from 10 up to +100 metre
    between 10 and 50 km, increasing to +300 metre at 100 km, after which
    the bias changes sign, crosses 0 at 113 km and finally reaches -4000
    metre at 150 km.  This is not due to humidity.  Atmlabs pt2z version
    differs only 30 metre from mine.  In %, this error is below 0.3% up to
    100 km, then changes sign and reaching -3% at 150 km.  For many
    purposes this is good enough, though, and certainly better than
    p2z_oversimplified.

    :param array p: Pressure [Pa]
    :param array T: Temperature [K].  Must match the size of p.
    :param h2o: Water vapour [vmr].  If negligible, set to 0.  Must be
        either scalar, or match the size of p and T.
    :param p0:
    :param z0:
    :param lat: Latitude [degrees].  This has some effect on the vertical
        distribution of gravitational acceleration, leading to difference
        of some 500 metre at 150 km.  Defaults to 45°.
    :param z_acc: Up to what precision to iteratively calculate the
        z-profile.  If -1, run two iterations, which should be accurate,
        according to the comment below.
    :param str ellps: Ellipsoid to use.  The function relies on
        pyproj.Geod, which is an interface to the proj library.  For a
        full table of ellipsoids, run 'proj -le'.
    :param bool extend: If p0, z0 outside of p, z range, extend
        artificially.  WARNING: This will assume CONSTANT T, h2o!
    :returns array z: Array of altitudes [m].  Same size as p and T.
"""
#     Original description:
# % PT2Z   Hydrostatic altitudes
# %
# %    Calculates altitudes fulfilling hydrostatic equilibrium, based on
# %    vertical profiles of pressure, temperature and water vapour. Pressure
# %    and altitude of a reference point must be specified.
# %
# %    Molecular weights and gravitational constants are hard coded and
# %    function is only valid for the Earth.
# %
# %    As the gravitation changes with altitude, an iterative process is
# %    needed. The accuracy can be controlled by *z_acc*. The calculations
# %    are repeated until the max change of the altitudes is below *z_acc*. If
# %    z_acc<0, the calculations are run twice, which should give an accuracy
# %    better than 1 m.
# %
# % FORMAT   z = pt2z( p, t, h2o, p0, z0 [,lat,z_acc,refell] )
# %       
# % OUT   z         Altitudes [m].
# % IN    p         Column vector of pressures [Pa].
# %       t         Column vector of temperatures [K].
# %       h2o       Water vapour [VMR]. Vector or a scalar, e.g. 0.
# %       p0        Pressure of reference point [Pa].
# %       z0        Altitude of reference point [m].
# %       lat       Latitude. Default is 45.
# %       z_acc     Accuracy for z. Default is -1.
# %       ellipsoid Reference ellipsoid data, see *ellipsoidmodels*.
# %                 Default is data matching WGS84.
# 
# % 2005-05-11   Created by Patrick Eriksson.

#32  function z = pt2z(p,t,h2o,p0,z0,varargin)
#33  %
#34  [lat,z_acc,ellipsoid] = optargs( varargin, { 45, -1, NaN } );
#35  %
    ellipsoid = pyproj.Geod(ellps=ellps)
#36  if isnan(ellipsoid)
#37    ellipsoid = ellipsoidmodels('wgs84');
#38  end
#39                                                                              %&%
#40  rqre_nargin( 5, nargin );                                                   %&%
#41  rqre_datatype( p, @istensor1 );                                             %&%
#42  rqre_datatype( t, @istensor1 );                                             %&%
#43  rqre_datatype( h2o, @istensor1 );                                           %&%
#44  rqre_datatype( p0, @istensor0 );                                            %&%
#45  rqre_datatype( z0, @istensor0 );                                            %&%
#46  rqre_datatype( lat, @istensor0 );                                           %&%

    if not p.size == T.size:
        raise ValueError("p and T must have same length")

    if p.min() < 0:
        raise ValueError("Found negative pressures")

    if T.min() < 0:
        raise ValueError("Found negative temperatures")

#47  np = length( p );
#48  if length(t) ~= np                                                          %&%
#49    error('The length of *p* and *t* must be identical.');                    %&%
#50  end                                                                         %&%

    if not (isinstance(h2o, numbers.Real) or h2o.size in (p.size, 1)):
        raise ValueError("h2o must have length of p or be scalar")

#51  if ~( length(h2o) == np  |  length(h2o) == 1 )                              %&%
#52    error('The length of *h2o* must be 1 or match *p*.');                     %&%
#53  end                                                                         %&%


# FIXME IS THIS NEEDED?  Yes — See e-mail Patrick 2014-08-11
    if p0 > p[0] or p0 < p[-1]:
        if extend:
            if p0 > p[0]: # p[0] is largest pressure, p0 even larger
                extend = "below"
                p = numpy.hstack([p0, p])
                T = numpy.hstack([T[0], T])
                h2o = numpy.hstack([h2o[0], h2o])
            elif p0 < p[-1]:
                extend = "above" # p[-1] is smallest pressure, p0 even smaller
                p = numpy.hstack([p, p0])
                T = numpy.hstack([T, T[-1]])
                h2o = numpy.hstack([h2o, h2o[-1]])
        else:
            raise ValueError(("reference pressure ({:.2f}) must be "
           "in total pressure range ({:.2f} -- {:.2f})").format(
               p0, p[0], p[-1]))
# END FIXME

#54  if p0 > p(1)  |  p0 < p(np)                                                 %&%
#55    error('Reference point (p0) can not be outside range of *p*.');           %&%
#56  end                                                                         %&%
#57  
#58  
#59  %= Expand *h2o* if necessary
#60  %
#61  if  length(h2o) == 1
#62    h2o = repmat( h2o, np, 1 );
#63  end
    if isinstance(h2o, numbers.Real) or h2o.size == 1:
        h2o = h2o * numpy.ones_like(p)

    if h2o.max() > 1:
        raise ValueError("Found h2o vmr values up to {:.2f}.  Expected < 1.".format(h2o.max()))
##64  
#65  
#66  %= Make rough estimate of *z*
#67  %
#68  z = p2z_simple( p );
    z = p2z_oversimplified(p)
#69  z = shift2refpoint( p, z, p0, z0 );
    z = _shift2refpoint(p, z, p0, z0)
#70  
#71  
#72  %= Set Earth radius and g at z=0
#73  %
#74  re = ellipsoidradii( ellipsoid, lat );
    # APPROXIMATION!  Approximate radius at latitude by linear
    # interpolation in cos(lat) between semi-major-axis and
    # semi-minor-axis
    #
    # Get radius at latitude
    re = (ellipsoid.a * numpy.cos(numpy.deg2rad(lat))
        + ellipsoid.b * (1-numpy.cos(numpy.deg2rad(lat))))
#75  g0 = lat2g0( lat );
    g0 = lat2g0(lat)
#76  
#77  
#78  %= Gas constant and molecular weight of dry air and water vapour
#79  %
#80  r  = constants( 'GAS_CONST' );
#81  md = 28.966;
#82  mw = 18.016;
#83  %
#84  k  = 1-mw/md;        % 1 - eps         
    k = 1 - M_w/M_d
#85  rd = 1e3 * r / md;   % Gas constant for 1 kg dry air
    rd = 1e3 * R / M_d  # gas constant for 1 kg dry air
#86  
#87  
#88  %= How to end iterations
#89  %
#90  if z_acc < 0
#91    niter = 2;
#92  else
#93    niter = 99;
#94  end 
    niter = 2 if z_acc < 0 else 99
#95  
#96  for iter = 1:niter
    for i in range(niter):
#97  
#98    zold = z;
#99   
        zold = z
#100   g = z2g( re, g0, z );
        g = z2g(re, g0, z)
#101 
#102   for i = 1 : (np-1)
        for i in range(p.size-1):
#103      
#104         gp  = ( g(i) + g(i+1) ) / 2;
            gp = (g[i] + g[i+1]) / 2
#105  
#106         %-- Calculate average water VMR (= average e/p)
#107         hm  = (h2o(i)+h2o(i+1)) / 2;
            hm = (h2o[i] + h2o[i+1]) / 2
#108  
#109         %--  The virtual temperature (no liquid water)
#110         tv = (t(i)+t(i+1)) / ( 2 * (1-hm*k) );   % E.g. 3.16 in Wallace&Hobbs
#111  
            tv = (T[i] + T[i+1]) / (2 * (1 - hm*k))

#112         %-- The change in vertical altitude from i to i+1
#113         dz = rd * (tv/gp) * log( p(i)/p(i+1) );
            dz = rd * (tv/gp) * numpy.log(p[i]/p[i+1])
#114         z(i+1) = z(i) + dz;
            z[i+1] = z[i] + dz
#115      
#116   end
#117  
#118   %-- Match the altitude of the reference point
#119   z = shift2refpoint( p, z, p0, z0 );
        z = _shift2refpoint(p, z, p0, z0)
#120 
#121   if z_acc >= 0 & max(abs(z-zold)) < z_acc
#122     break;
#123   end
        if z_acc >= 0 and max(abs(z-zold)) < z_acc:
            break
#124  
#125 end
#126 
#127 return
    # correct for extending
    if extend == "below": # lowest pressure extra
        return z[1:]
    elif extend == "above": # highest pressure extra
        return z[:-1]
    else:
        return z

#128 %----------------------------------------------------------------------------
#129 
#130 function z = shift2refpoint( p, z, p0, z0 )
#131   %
#132   z = z - ( interpp( p, z, p0 ) - z0 );
#133   %
#134 return

def _shift2refpoint(p, z, p0, z0):
    """Given z(p), shift this to include (p0, z0)

    Taken from atmlabs equivalent function
    https://www.sat.ltu.se/trac/rt/browser/atmlab/trunk/geophysics/pt2z.m
    """
    #return z - (pamath.interpp(p, z, p0) - z0)
    # revert p, z because for numpy.interp x-coor must be increasing
    return z - (numpy.interp(numpy.log(p0), numpy.log(p[::-1]), z[::-1]) - z0)

def z2g(r_geoid, g0, z):
    """Calculate gravitational acceleration at elevation

    Derived from atmlabs equivalent function
    https://www.sat.ltu.se/trac/rt/browser/atmlab/trunk/geophysics/pt2z.m

    :param r: surface radius at point [m]
    :param g0: surface gravitational acceleration at point [m/s^2]
    :param z: elevation [m]
    :returns: gravitational acceleration at point [m/s^2]
    """
#137 function g = z2g(r_geoid,g0,z)
#138   %
#139   g = g0 * (r_geoid./(r_geoid+z)).^2;
    return g0 * (r_geoid/(r_geoid+z))**2;
#140   %
#141 return
#142 

def lat2g0(lat):
    """Calculate surface gravitational acceleration for latitude

    This function is stolen from atmlab:
    https://www.sat.ltu.se/trac/rt/browser/atmlab/trunk/geophysics/pt2z.m

    From the original description:

    Expression below taken from Wikipedia page "Gravity of Earth", that is stated
    to be: International Gravity Formula 1967, the 1967 Geodetic Reference System
    Formula, Helmert's equation or Clairault's formula.

    :param lat: Latitude [degrees]
    :returns: gravitational acceleration [m/s]
    """

    x  = numpy.abs( lat );
    # see docstring for source of parametrisation
    return 9.780327 * ( 1 + 5.3024e-3*numpy.sin(numpy.deg2rad(x))**2 
                          + 5.8e-6*numpy.sin(numpy.deg2rad(2*x)**2 ))


def estimate_effective_temperature(f, W, f_c, T):
    r"""Estimate effective temperature for SRF

    For a monochromatic radiance, one can easily convert radiance to
    brightness temperature using the Planck function.  For a polychromatic
    radiance such as a channel radiance, this is incorrect.  Weinreb et
    al. (1981) propose a solution in calculating an effective temperature:

    T_e = B^{-1}(\int_{f_1}^{f_2} \phi B df)

    where $T_e$ is the effective temperature, $B$ is the Planck function,
    $f_1$ and $f_2$ are the lower and upper limit of the channel, $\phi$
    is the normalised channel radiance.  This, used with the central
    wavenumber, is correct FIXME VERIFY.  According to Weinreb et al.
    (1981) this is "from the theorem"...  I should understand this!
    
    Weinreb, M.P., Fleming, H.E., McMillin, L.M., Neuendorffer, A.C,
    Transmittances for the TIROS operational vertical sounder,
    1981, NOAA Technical Report NESS 85.

    :param ndarray f: Frequency [Hz] for spectral response function.
    :param ndarray W: Weight [1] for spectral response function.  Should
        be normalised to be integrated to 1.
    :param f_c: Central frequency to be used with Planck function.
    :param T: Temperature [K]
    :returns: Effective temperature [K].
    """

    B = planck_f(f, T)
    chanrad = spectral_to_channel(f, B, f.squeeze(), W)
    T_e = specrad_frequency_to_planck_bt(chanrad, f_c)
    return T_e



"""Datasets for TOVS/ATOVS
"""

import io
import tempfile
import subprocess
import datetime
import logging
import gzip
import shutil
import abc
import pathlib
import dbm

import numpy
import scipy.interpolate

import netCDF4
import dateutil
import progressbar

try:
    import coda
except ImportError:
    logging.warn("Unable to import coda, won't read IASI EPS L1C")
    
import typhon.datasets.dataset
import typhon.utils.metaclass
import typhon.physics.units
from typhon.datasets.tovs import (Radiometer, HIRS, HIRSPOD, HIRS2,
    HIRSKLM, HIRS3, HIRS4)

from .. import dataset
from .. import tools
from .. import constants
from .. import physics
from .. import math as pamath
from ..units import ureg
from .. import config
from .. import units

from . import _tovs_defs

    
class HIRS2I(HIRS2):
    # identical fileformat, I believe
    satellites = {"noaa11", "noaa14"}

# HIRSFCDR and friends now in FCDR_HIRS.fcdr

class IASINC(dataset.MultiFileDataset, typhon.datasets.dataset.HyperSpectral):
    """Read IASI from NetCDF
    """
    _dtype = numpy.dtype([
        ("time", "M8[s]"),
        ("lat", "f4"),
        ("lon", "f4"),
        ("satellite_zenith_angle", "f4"),
        ("satellite_azimuth_angle", "f4"),
        ("solar_zenith_angle", "f4"),
        ("solar_azimuth_angle", "f4"),
        ("spectral_radiance", "f4", 8700)])
    name = "iasinc"
    start_date = datetime.datetime(2003, 1, 1, 0, 0, 0)
    end_date = datetime.datetime(2013, 12, 31, 23, 59, 59)
    granule_duration = datetime.timedelta(seconds=1200)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        freqfile = self.basedir / "frequency.txt"
        if freqfile.exists():
            self.frequency = numpy.loadtxt(str(freqfile))
            
    def _read(self, path, fields="all", return_header=False):
        if fields == "all":
            fields = self._dtype.names
        logging.debug("Reading {!s}".format(path))
        with netCDF4.Dataset(str(path), 'r', clobber=False) as ds:
            scale = ds["scale_factor"][:]
            scale_valid = numpy.isfinite(scale) & (scale > 0)
            wavenumber = ureg.Quantity(ds["wavenumber"][:],
                ureg.parse_expression(ds["wavenumber"].units.replace("m-1", "m^-1")))
            wavenumber_valid = numpy.isfinite(wavenumber) & (wavenumber.m > 0)
            if not numpy.array_equal(scale_valid, wavenumber_valid):
                raise ValueError("Scale and wavenumber inconsistently valid")
            if self.wavenumber is None:
                self.wavenumber = wavenumber[wavenumber_valid]
            elif (abs(self.wavenumber - wavenumber[wavenumber_valid]).max()
                    > (0.05 * (1/ureg.centimetre))):
                raise ValueError("Inconsistent wavenumbers!")

            dtp = [x for x in self._dtype.descr if x[0] in fields]
            if dtp[-1][0] == "spectral_radiance":
                dtp[-1] = (dtp[-1][0], dtp[-1][1], wavenumber_valid.sum())

            M = numpy.zeros(
                dtype=dtp,
                shape=(len(ds.dimensions["along_track"]),
                       len(ds.dimensions["across_track"])))
            time_ref = numpy.datetime64(datetime.datetime.strptime(
                        ds["time"].gsics_reference_time,
                        "%Y-%m-%dT%H:%M:%S+00:00"), "s")
            dlt = numpy.array(ds["time"][:], dtype="m8[s]")
            M["time"] = (time_ref + dlt)[:, numpy.newaxis]
            for var in set(M.dtype.names) - {"time", "spectral_radiance"}:
                M[var] = ds[var][...]
            if "spectral_radiance" in M.dtype.names:
                M["spectral_radiance"][:, :, :] = (
                        ds["spectral_radiance"][:, :, scale_valid] /
                        scale[scale_valid][numpy.newaxis, numpy.newaxis, :])

        return M

class IASIEPS(dataset.MultiFileDataset, typhon.datasets.dataset.HyperSpectral):
    """Read IASI from EUMETSAT EPS L1C
    """

    name = "iasi"
    start_date = datetime.datetime(2007, 5,  29, 5, 8, 56)
    end_date = datetime.datetime(2015, 11, 17, 16, 38, 59)
    granule_duration = datetime.timedelta(seconds=6200)
    _dtype = numpy.dtype([
        ("time", "M8[ms]"),
        ("lat", "f4", (4,)),
        ("lon", "f4", (4,)),
        ("satellite_zenith_angle", "f4", (4,)),
        ("satellite_azimuth_angle", "f4", (4,)),
        ("solar_zenith_angle", "f4", (4,)),
        ("solar_azimuth_angle", "f4", (4,)),
        ("spectral_radiance", "f4", (4, 8700))])

    # Minimum temporary space for unpacking
    # Warning: race conditions needs to be addressed.
    # As a workaround, choose very large minspace.
    minspace = 1e10

    @staticmethod
    def __obtain_from_mdr(c, field):
        fieldall = numpy.concatenate([getattr(x.MDR, field)[:, :, :,
            numpy.newaxis] for x in c.MDR if hasattr(x, 'MDR')], 3)
        fieldall = numpy.transpose(fieldall, [3, 0, 1, 2])
        return fieldall

    def _read(self, path, fields="all", return_header=False):
        tmpdira = config.conf["main"]["tmpdir"]
        tmpdirb = config.conf["main"]["tmpdirb"]
        tmpdir = (tmpdira 
            if shutil.disk_usage(tmpdira).free > self.minspace
            else tmpdirb)
            
        with tempfile.NamedTemporaryFile(mode="wb", dir=tmpdir, delete=True) as tmpfile:
            with gzip.open(str(path), "rb") as gzfile:
                logging.debug("Decompressing {!s}".format(path))
                gzcont = gzfile.read()
                logging.debug("Writing decompressed file to {!s}".format(tmpfile.name))
                tmpfile.write(gzcont)
                del gzcont

            # All the hard work is in coda
            logging.debug("Reading {!s}".format(tmpfile.name))
            cfp = coda.open(tmpfile.name)
            c = coda.fetch(cfp)
            logging.debug("Sorting info...")
            n_scanlines = c.MPHR.TOTAL_MDR
            start = datetime.datetime(*coda.time_double_to_parts_utc(c.MPHR.SENSING_START))
            has_mdr = numpy.array([hasattr(m, 'MDR') for m in c.MDR],
                        dtype=numpy.bool)
            bad = numpy.array([
                (m.MDR.DEGRADED_PROC_MDR|m.MDR.DEGRADED_INST_MDR)
                        if hasattr(m, 'MDR') else True
                        for m in c.MDR],
                            dtype=numpy.bool)
            dlt = numpy.concatenate(
                [m.MDR.OnboardUTC[:, numpy.newaxis]
                    for m in c.MDR
                    if hasattr(m, 'MDR')], 1) - c.MPHR.SENSING_START
            M = numpy.ma.zeros(
                dtype=self._dtype,
                shape=(n_scanlines, 30))
            M["time"][has_mdr] = numpy.datetime64(start, "ms") + numpy.array(dlt*1e3, "m8[ms]").T
            specall = self.__obtain_from_mdr(c, "GS1cSpect").astype("f8")
            # apply scale factors
            first = c.MDR[0].MDR.IDefNsfirst1b
            last = c.MDR[0].MDR.IDefNslast1b
            for (slc_st, slc_fi, fact) in zip(
                    filter(None, c.GIADR_ScaleFactors.IDefScaleSondNsfirst),
                    c.GIADR_ScaleFactors.IDefScaleSondNslast,
                    c.GIADR_ScaleFactors.IDefScaleSondScaleFactor):
                # Documented intervals are closed [a, b]; Python uses
                # half-open [a, b).
                specall[..., (slc_st-first):(slc_fi-first+1)] *= pow(10.0, -fact)
            M["spectral_radiance"][has_mdr] = specall
            locall = self.__obtain_from_mdr(c, "GGeoSondLoc")
            M["lon"][has_mdr] = locall[:, :, :, 0]
            M["lat"][has_mdr] = locall[:, :, :, 1]
            satangall = self.__obtain_from_mdr(c, "GGeoSondAnglesMETOP")
            M["satellite_zenith_angle"][has_mdr] = satangall[:, :, :, 0]
            M["satellite_azimuth_angle"][has_mdr] = satangall[:, :, :, 1]
            solangall = self.__obtain_from_mdr(c, "GGeoSondAnglesSUN")
            M["solar_zenith_angle"][has_mdr] = solangall[:, :, :, 0]
            M["solar_azimuth_angle"][has_mdr] = solangall[:, :, :, 1]
            for fld in M.dtype.names:
                M.mask[fld][~has_mdr, ...] = True
                M.mask[fld][bad, ...] = True
            m = c.MDR[0].MDR
            wavenumber = (m.IDefSpectDWn1b * numpy.arange(m.IDefNsfirst1b, m.IDefNslast1b+0.1) * (1/ureg.metre))
            if self.wavenumber is None:
                self.wavenumber = wavenumber
            elif abs(self.wavenumber - wavenumber).max() > (0.05 * 1/(ureg.centimetre)):
                raise ValueError("Inconsistent wavenumbers")
            return M

class IASISub(dataset.HomemadeDataset, typhon.datasets.dataset.HyperSpectral):
    name = "iasisub"
    subdir = "{month}"
    stored_name = "IASI_1C_selection_{year}_{month}_{day}.npz"
    re = r"IASI_1C_selection_(?P<year>\d{4})_(?P<month>\d{1,2})_(?P<day>\d{1,2}).npz"
    start_date = datetime.datetime(2011, 1, 1, 0, 0, 0)
    end_date = datetime.datetime(2011, 12, 31, 23, 59, 59)

    
    def _read(self, *args, **kwargs):
        if self.frequency is None:
            self.frequency = numpy.loadtxt(self.freqfile)
        return super()._read(*args, **kwargs)

    def get_times_for_granule(self, p, **kwargs):
        gd = self.get_info_for_granule(p)
        (year, month, day) = (int(gd[m]) for m in "year month day".split())
        # FIXME: this isn't accurate, it usually starts slightly later...
        start = datetime.datetime(year, month, day, 0, 0, 0)
        # FIXME: this isn't accurate, there may be some in the next day...
        end = datetime.datetime(year, month, day, 23, 59, 59)
        return (start, end)



def which_hirs_fcdr(satname):
    """Given a satellite, return right HIRS object
    """
    for h in {HIRS2FCDR, HIRS3FCDR, HIRS4FCDR}:
        if satname in h.satellites:
            return h()
            break
    else:
        raise ValueError("Unknown HIRS satellite: {:s}".format(satname))



#!/usr/bin/env python
# coding: utf-8

"""Various statistical functions

"""

import numpy
import matplotlib.mlab

from typhon.math.stats import (bin, bin_nd)

def bin_nd_sparse(binners, bins):
    """Similar to bin_nd, but for very sparse data

    When the number of entries to be binned is much smaller than the
    number of bins, it is very inefficient to go through all the bins as
    bin_nd does.  This function takes the same arguments but instead of
    returning an n-dimensional array, it will return an n*p array where n
    are the number of dimensions and p the number of datapoints, returning
    the coordinates that each point would be in.
    """

    # FIXME: len(binners), or binners.shape[0]?
    M = numpy.zeros(shape=(len(bins), len(binners)), dtype="u2")
    for i in range(len(bins)):
        binner = binners[:, i]
        bin = bins[i]
        if numpy.isnan(binner).any():
            raise ValueError("I found nans in bin {:d}.  I refuse to bin nans!".format(
                i))
        M[i, :] = numpy.digitize(binner, bin)
    return M
        

def binsnD_to_2d(bins, ax1, ax2):
    """Collapse n-D bucketing array to 2-D, merging buckets for other dims.

    Take a n-D bucketing array (n>=2), such as returned by bin_nD, and
    merge all buckets thus reducing it to 2D, retaining ax1 and ax2.
    For examble, binsnD_flatiter(bins, 1, 2) will return a bins-array of
    shape (bins.shape[1], bins.shape[2]), where each bucket [i, j] will
    contain all elements in bins[:, i, j, ...].

    :param bins: n-Dimensional array containing buckets.  As returned by
        bin_nD.
    :param ax1: First axis to keep.
    :param ax2: Second axis to keep.
    :returns: 2-Dimensional array of shape (bins.shape[ax1],
        bins.shape[ax2]) with all other dimensions merged.
    """
    slc = [slice(None)] * bins.ndim # NB: slice(None) corresponds to :
    Z = numpy.empty(shape=(bins.shape[ax1], bins.shape[ax2]),
                    dtype="object")
    for i in range(bins.shape[ax1]):
        for j in range(bins.shape[ax2]):
            slc[ax1] = i
            slc[ax2] = j
            Z[i, j] = numpy.concatenate(bins[slc].ravel())

    return Z


class PCA(matplotlib.mlab.PCA):
    """Extend mlabs PCA class with a bit more functionality
    """

    def reconstruct(self, Y, truncate=0):
        """Reconstruct original vector from PC vector

        From a vector in PC axes, reconstruct the original vector.

        :param pcdp: L×n vector in PC space
        :param denormalise: 
        """

        truncate = truncate or self.Wt.shape[0]
        yscld = self.Wt[:truncate, :] @ Y[:, :truncate]
        return self.decenter(yscld)

    def decenter(self, yscld):
        """Undo self.center
        """
        return yscld * self.sigma + self.mu

    def estimate(self, y, n):
        """Estimate missing value

        Estimate the missing element for a vector.  For example, if twelve
        channels were used to construct a PCA but we have only 11, we can
        estimate what the 12th would have been, assuming we know which one
        is missing, of course.

        :param y: ndarray of dimensions L × n, where L can be any number
            and n should match the number of PCs.  The element that is to
            be predicted is thus contained in this array (probably masked)
            but not used by this method.
        :param n: index of which element is to be estimated.
        :returns: Full vector of same dimension as y, but with missing
            channel replace by a new estimate.
        """

        idx = numpy.delete(numpy.arange(y.shape[1]), n)

        yscld = self.center(y)

        est = self.decenter((self.Wt[:, idx] @ yscld[:, idx].T).T @ self.Wt)

        return est



"""This package contains modules defining datasets
"""

# imports for backward compatibility

from .groundbased import (Eureka_PRL_CH4_HDF, NDACCAmes)

from .tanso import (TansoFTSBase, TansoFTSv10x, TansoFTSv001)

from .acefts import ACEFTS



#!/usr/bin/env python
# coding: utf-8

"""Interact with matplotlib and other plotters

"""

import os.path
import datetime
now = datetime.datetime.now
import logging
import subprocess
import sys
import pickle
import lzma
import pathlib

import numpy
import matplotlib
import matplotlib.pyplot
import mpl_toolkits.basemap
from . import config
from . import io
from . import meta
from . import tools

def pcolor_on_map(m, lon, lat, C, **kwargs):
    """Wrapper around pcolor on a map, in case we cross the IDL

    preventing spurious lines
    """

    # Need to investigate why and how to solve:
    # -175 - minor polar problems (5° missing)
    # -178 - polar problems (2° missing)
    # -179 - some problems (1° missing)
    # -179.9 - many problems
    # perhaps I need to mask neighbours or so?
    C1 = numpy.ma.masked_where(lon<-175, C, copy=True)
    p1 = m.pcolor(lon, lat, C1, latlon=True, **kwargs)
#    C2 = numpy.ma.masked_where(lon<0, C.data)
#    p2 = m.pcolor(lon, lat, C2, latlon=True, **kwargs)
#    mixed = lon.ptp(1)>300
#    homog = ~mixed
#    homognum = homog.nonzero()[0]
#
#    breaks = numpy.diff(homognum) > 5
#    breakedge = numpy.r_[-1, breaks.nonzero()[0], homog.sum()-1]
#
#    for (l, e) in zip(breakedge, breakedge[1:]):
#        p1 = m.pcolor(lon[homognum[(l+1):(e+1)], :],
#                      lat[homognum[(l+1):(e+1)], :],
#                      C[homognum[(l+1):(e+1)], :],
#                      latlon=True, **kwargs)
#    lon.mask = lat.mask = C.mask = west
#    p2 = m.pcolor(lon, lat, C, latlon=True, **kwargs)
    # remaining lines manually
#    for mix in mixed.nonzero()[0]:
#        west = lon[mix, :] <= 0
#        east = lon[mix, :] > 0
#        for h in (west, east):
#            m.pcolor(lon[mix, h],
#                     lat[mix, h],
#                     C[mix, h],
#                     latlon=True, **kwargs)
            # For some reason, if I don't include ascontiguousarray here,
            # I run into a SystemError in proj4.  I haven't been able to
            # find a minimum example that reproduces the bug :(
            #
            # And another bug: I can't pcolor a single line when using
            # latlon=True, as shiftdata will fail...
            #
            # But even when I can, it still goes wrong because pcolor
            # doesn't show the single line... :( why is masking not
            # working?

#            (x, y) = m(numpy.ascontiguousarray(lon[mix:(mix+1), h]),
#                       numpy.ascontiguousarray(lat[mix:(mix+1), h]))
#            m.pcolor(x, y, C[mix:(mix+1), h], latlon=False, **kwargs)
    return p1

def map_orbit_double_with_stats(lon, lat, C, U, lab1, lab2,  title, filename):
    """Map orbit with uncertainty and histograms
    """

    (f, a_all) = matplotlib.pyplot.subplots(2, 4,
                gridspec_kw = {'width_ratios':[12, 1, 3, 8],
                               "hspace": 0.3},
                figsize=(15, 8))

    # workaround for easy way of creating extra space...
    for a in a_all[:, 2]:
        a.set_visible(False)

    m_all = []
    for a in a_all[:, 0]:
        m = mpl_toolkits.basemap.Basemap(projection="moll",
                resolution="c", ax=a, lon_0=0)
        m.drawcoastlines()
        m.drawmeridians(numpy.arange(-180, 180, 30))
        m.drawparallels(numpy.arange(-90, 90, 30))
        m_all.append(m)

    pcr = pcolor_on_map(
        m_all[0], lon, lat,
        C, cmap="viridis")

    pcu = pcolor_on_map(
        m_all[1], lon, lat,
        U, cmap="inferno_r")

    cb1 = f.colorbar(pcr, cax=a_all[0, 1])
    cb1.set_label(lab1)#"Counts")

    cb2 = f.colorbar(pcu, cax=a_all[1, 1])
    cb2.set_label(lab2)#"Random uncertainty [counts]")

    a_all[0, 3].hist(C.ravel(), 50)
    a_all[0, 3].set_xlabel(lab1)#"Counts")
    a_all[1, 3].hist(U.ravel(), 50)
    a_all[1, 3].set_xlabel(lab2)#r"$\Delta$ Counts")
    for a in a_all[:, 3]:
        a.grid("on")
        a.set_ylabel("Number")

    f.suptitle(title)

    #f.subplots_adjust(wspace=0.2, hspace=0.2)

    print_or_show(f, False, filename)

def plotdir():
    """Returns todays plotdir.

    Configuration 'plotdir' must be set.  Value is expanded with strftime.
    """
    return datetime.date.today().strftime(config.get_config('plotdir'))

def print_or_show(fig, show, outfile, in_plotdir=True, tikz=None,
                  data=None, store_meta="", close=True,
                  dump_pickle=True):
    """Either print or save figure, or both, depending on arguments.

    Taking a figure, show and/or save figure in the default directory,
    obtained with :func:plotdir.  Creates plot directory if needed.

    :param fig: Figure to store.  
    :type fig: matplotlib.Figure object
    :param show: Show figure or not
    :type show: boolean
    :param outfile: File to write figure to, or list of files.  If the
        string ends in a '.', write to x.png and x.pdf.
    :type outfile: string or list of strings
    :param in_plotdir: If true, write to default plot directory.  If
        false, write to currect directory or use absolute path.
    :type in_plotdir: boolean
    :param tikz: Try to write tikz code with matplotlib2tikz.  Requires
        that the latter is installed.
    :type tikz: boolean
    :param data: Store associated data in .dat file (useful for pgfplots).
        May be a list of ndarrays, which results in multiple numbered datafiles.
    :type data: ndarray or list thereof
    :param store_meta: Also store other info.  This is a string that will
        be written to a file.  If not set or set to None, it will just
        write the pyatmlab version.  The file will use the same basename
        as the outfile, but replacing the extention by "info".  However,
        this only works if outfile is a string and not a list thereof.
        To write nothing, pass an empty string.
    :type store_meta: str.
    :param close: If true, close figure.  Defaults to true.
    :type close: bool.
    """

    if outfile is not None:
        outfiles = [outfile] if isinstance(outfile, str) else outfile
        
        bs = pathlib.Path(plotdir())
        if isinstance(outfile, str):
            if outfile.endswith("."):
                outfiles = [bs / pathlib.Path(outfile+ext) for ext in ("png", "pdf")]
                infofile = bs / pathlib.Path(outfile + "info")
                figfile = bs / pathlib.Path(outfile + "pkl.xz")
            else:
                outfiles = [bs / pathlib.Path(outfile)]
                infofile = None
                figfile = None

        if infofile is not None:
            infofile.parent.mkdir(parents=True, exist_ok=True)

            logging.debug("Obtaining verbose stack info")
            pr = subprocess.run(["pip", "freeze"], stdout=subprocess.PIPE) 
            info = " ".join(sys.argv) + "\n" + pr.stdout.decode("utf-8") + "\n"
            info += tools.get_verbose_stack_description()

#        if infofile is not None and info:
            logging.info("Writing info to {!s}".format(infofile))
            with infofile.open("w", encoding="utf-8") as fp:
                fp.write(info)
        if dump_pickle and figfile is not None:
            logging.info("Writing figure object to {!s}".format(figfile))
            with lzma.open(str(figfile), "wb", preset=lzma.PRESET_DEFAULT) as fp:
                pickle.dump(fig, fp, protocol=4)
        # interpret as sequence
        for outf in outfiles:
            logging.info("Writing to file: {!s}".format(outf))
            outf.parent.mkdir(parents=True, exist_ok=True)
            fig.canvas.print_figure(str(outf))
    if show:
        matplotlib.pyplot.show()

    if close:
        matplotlib.pyplot.close(fig)

    if tikz is not None:
        import matplotlib2tikz
        print(now(), "Writing also to:", os.path.join(plotdir(), tikz))
        matplotlib2tikz.save(os.path.join(plotdir(), tikz))
    if data is not None:
        if not os.path.exists(io.plotdatadir()):
            os.makedirs(io.plotdatadir())
        if isinstance(data, numpy.ndarray):
            data = (data,)
        # now take it as a loop
        for (i, dat) in enumerate(data):
            outf = os.path.join(io.plotdatadir(),
                "{:s}{:d}.dat".format(
                    os.path.splitext(outfiles[0])[0], i))
            fmt = ("%d" if issubclass(dat.dtype.type, numpy.integer) else
                    '%.18e')
            if len(dat.shape) < 3:
                numpy.savetxt(outf, dat, fmt=fmt)
            elif len(dat.shape) == 3:
                io.savetxt_3d(outf, dat, fmt=fmt)
            else:
                raise ValueError("Cannot write {:d}-dim ndarray to textfile".format(
                    len(dat.shape)))
